import json
import os
import textwrap
import unittest

from parameterized import parameterized
import pytest

from conans.model.graph_lock import LOCKFILE
from conans.test.assets.genconanfile import GenConanfile
from conans.test.utils.tools import TestClient, TestServer
from conans.util.env_reader import get_env
from conans.util.files import load


conanfile = textwrap.dedent("""
    from conans import ConanFile, load
    import os
    class Pkg(ConanFile):
        {requires}
        exports_sources = "myfile.txt"
        keep_imports = True
        def imports(self):
            self.copy("myfile.txt", folder=True)
        def package(self):
            self.copy("*myfile.txt")
        def package_info(self):
            self.output.info("SELF FILE: %s"
                % load(os.path.join(self.package_folder, "myfile.txt")))
            for d in os.listdir(self.package_folder):
                p = os.path.join(self.package_folder, d, "myfile.txt")
                if os.path.isfile(p):
                    self.output.info("DEP FILE %s: %s" % (d, load(p)))
        """)


class GraphLockCITest(unittest.TestCase):

    @parameterized.expand([("recipe_revision_mode",), ("package_revision_mode",)])
    @pytest.mark.skipif(not get_env("TESTING_REVISIONS_ENABLED", False), reason="Only revisions")
    def test_revisions(self, package_id_mode):
        test_server = TestServer(users={"user": "mypass"})
        client = TestClient(servers={"default": test_server},
                            users={"default": [("user", "mypass")]})
        client.run("config set general.default_package_id_mode=%s" % package_id_mode)
        client.save({"conanfile.py": conanfile.format(requires=""),
                     "myfile.txt": "HelloA"})
        client.run("create . PkgA/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(
            requires='requires = "PkgA/0.1@user/channel"'),
                     "myfile.txt": "HelloB"})
        client.run("create . PkgB/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(
            requires='requires = "PkgB/0.1@user/channel"'),
                     "myfile.txt": "HelloC"})
        client.run("create . PkgC/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(
            requires='requires = "PkgC/0.1@user/channel"'),
                     "myfile.txt": "HelloD"})
        client.run("create . PkgD/0.1@user/channel")
        self.assertIn("PkgD/0.1@user/channel: SELF FILE: HelloD", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgA: HelloA", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgC: HelloC", client.out)

        client.run("upload * --all --confirm")

        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        initial_lock_file = client.load(LOCKFILE)

        # Do a change in B, this will be a new revision
        clientb = TestClient(cache_folder=client.cache_folder, servers={"default": test_server})
        clientb.save({"conanfile.py": conanfile.format(requires='requires="PkgA/0.1@user/channel"'),
                     "myfile.txt": "ByeB World!!"})
        clientb.run("create . PkgB/0.1@user/channel")

        # Go back to main orchestrator
        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        client.run("lock build-order conan.lock --json=build_order.json")
        master_lockfile = client.load("conan.lock")

        build_order = client.load("build_order.json")
        to_build = json.loads(build_order)
        lock_fileaux = master_lockfile
        while to_build:
            for ref, _, _, _ in to_build[0]:
                client_aux = TestClient(cache_folder=client.cache_folder,
                                        servers={"default": test_server})
                client_aux.save({LOCKFILE: lock_fileaux})
                client_aux.run("install %s --build=%s --lockfile=conan.lock"
                               " --lockfile-out=conan.lock" % (ref, ref))
                lock_fileaux = load(os.path.join(client_aux.current_folder, LOCKFILE))
                client.save({"new_lock/%s" % LOCKFILE: lock_fileaux})
                client.run("lock update conan.lock new_lock/conan.lock")

            client.run("lock build-order conan.lock --json=bo.json")
            lock_fileaux = client.load(LOCKFILE)
            to_build = json.loads(client.load("bo.json"))

        new_lockfile = client.load(LOCKFILE)
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)
        client.run("upload * --all --confirm")

        client.save({LOCKFILE: initial_lock_file})
        client.run("remove * -f")
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)

        client.save({LOCKFILE: new_lockfile})
        client.run("remove * -f")
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)

    @parameterized.expand([(False,), (True,)])
    def test_version_ranges(self, partial_lock):
        client = TestClient()
        client.run("config set general.default_package_id_mode=full_package_mode")
        files = {
            "pkga/conanfile.py": conanfile.format(requires=""),
            "pkga/myfile.txt": "HelloA",
            "pkgb/conanfile.py": conanfile.format(requires='requires="PkgA/[*]@user/channel"'),
            "pkgb/myfile.txt": "HelloB",
            "pkgc/conanfile.py": conanfile.format(requires='requires="PkgB/[*]@user/channel"'),
            "pkgc/myfile.txt": "HelloC",
            "pkgd/conanfile.py": conanfile.format(requires='requires="PkgC/[*]@user/channel"'),
            "pkgd/myfile.txt": "HelloD",
        }
        client.save(files)

        client.run("create pkga PkgA/0.1@user/channel")
        client.run("create pkgb PkgB/0.1@user/channel")
        client.run("create pkgc PkgC/0.1@user/channel")
        client.run("create pkgd PkgD/0.1@user/channel")
        self.assertIn("PkgD/0.1@user/channel: SELF FILE: HelloD", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgA: HelloA", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgC: HelloC", client.out)

        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        initial_lockfile = client.load("conan.lock")

        # Do a change in B
        client.save({"pkgb/myfile.txt": "ByeB World!!"})
        if not partial_lock:
            client.run("export pkgb PkgB/0.2@user/channel")

            # Go back to main orchestrator
            client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=productd.lock")

            # Now it is locked, PkgA can change
            client.save({"pkga/myfile.txt": "ByeA World!!"})
            client.run("create pkga PkgA/0.2@user/channel")
        else:
            client.run("lock create pkgb/conanfile.py --name=PkgB --version=0.2 --user=user "
                       "--channel=channel --lockfile-out=buildb.lock")
            self.assertIn("PkgA/0.1", client.out)
            self.assertNotIn("PkgA/0.2", client.out)

            # Now it is locked, PkgA can change
            client.save({"pkga/myfile.txt": "ByeA World!!"})
            client.run("create pkga PkgA/0.2@user/channel")

            # Package can be created with previous lock, keep PkgA/0.1
            client.run("create pkgb PkgB/0.2@user/channel --lockfile=buildb.lock "
                       "--lockfile-out=buildb.lock")
            self.assertIn("PkgA/0.1", client.out)
            self.assertNotIn("PkgA/0.2", client.out)
            self.assertIn("PkgB/0.2@user/channel: DEP FILE PkgA: HelloA", client.out)
            self.assertNotIn("ByeA", client.out)
            buildblock = client.load("buildb.lock")

            # Go back to main orchestrator, buildb.lock can be used to lock PkgA/0.1 too
            client.save({"buildb.lock": buildblock})
            client.run("lock create --reference=PkgD/0.1@user/channel --lockfile=buildb.lock "
                       "--lockfile-out=productd.lock")
            self.assertIn("PkgA/0.1", client.out)
            self.assertNotIn("PkgA/0.2", client.out)

        client.run("lock build-order productd.lock --json=build_order.json")
        productd_lockfile = client.load("productd.lock")

        json_file = client.load("build_order.json")
        to_build = json.loads(json_file)
        lock_fileaux = productd_lockfile
        while to_build:
            for ref, _, _, _ in to_build[0]:
                client_aux = TestClient(cache_folder=client.cache_folder)
                client_aux.save({"productd.lock": lock_fileaux})
                client_aux.run("install %s --build=%s --lockfile=productd.lock "
                               "--lockfile-out=productd.lock" % (ref, ref))
                lock_fileaux = client_aux.load("productd.lock")
                client.save({"new_lock/productd.lock": lock_fileaux})
                client.run("lock update productd.lock new_lock/productd.lock")

            client.run("lock build-order productd.lock --json=bo.json")
            lock_fileaux = client.load("productd.lock")
            to_build = json.loads(client.load("bo.json"))

        # Make sure built packages are marked as modified
        productd_lockfile = client.load("productd.lock")
        productd_lockfile_json = json.loads(productd_lockfile)
        nodes = productd_lockfile_json["graph_lock"]["nodes"]
        pkgb = nodes["0"] if partial_lock else nodes["3"]
        pkgc = nodes["4"] if partial_lock else nodes["2"]
        pkgd = nodes["3"] if partial_lock else nodes["1"]
        self.assertIn("PkgB/0.2", pkgb["ref"])
        self.assertTrue(pkgb["modified"])
        self.assertIn("PkgC/0.1", pkgc["ref"])
        self.assertTrue(pkgc["modified"])
        self.assertIn("PkgD/0.1", pkgd["ref"])
        self.assertTrue(pkgd["modified"])

        new_lockfile = client.load("productd.lock")
        client.run("install PkgD/0.1@user/channel --lockfile=productd.lock")
        self.assertIn("HelloA", client.out)
        self.assertNotIn("ByeA", client.out)
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)

        client.save({LOCKFILE: initial_lockfile})
        self.assertIn("HelloA", client.out)
        self.assertNotIn("ByeA", client.out)
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)

        client.save({LOCKFILE: new_lockfile})
        self.assertIn("HelloA", client.out)
        self.assertNotIn("ByeA", client.out)
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: ByeB World!!", client.out)

        # Not locked will retrieve newer versions
        client.run("install PkgD/0.1@user/channel", assert_error=True)
        self.assertIn("PkgA/0.2@user/channel:5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9 - Cache",
                      client.out)
        self.assertIn("PkgB/0.2@user/channel:11b376c6e7a22ec390c215a8584ef9237a6da32f - Missing",
                      client.out)

    def test_version_ranges_diamond(self):
        client = TestClient()
        client.run("config set general.default_package_id_mode=full_package_mode")
        client.save({"conanfile.py": conanfile.format(requires=""),
                     "myfile.txt": "HelloA"})
        client.run("create . PkgA/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(requires='requires="PkgA/[*]@user/channel"'),
                     "myfile.txt": "HelloB"})
        client.run("create . PkgB/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(requires='requires="PkgA/[*]@user/channel"'),
                     "myfile.txt": "HelloC"})
        client.run("create . PkgC/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(requires='requires="PkgB/[*]@user/channel",'
                                                      ' "PkgC/[*]@user/channel"'),
                     "myfile.txt": "HelloD"})
        client.run("create . PkgD/0.1@user/channel")
        self.assertIn("PkgD/0.1@user/channel: SELF FILE: HelloD", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgA: HelloA", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgC: HelloC", client.out)

        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        lock_file = client.load(LOCKFILE)
        initial_lock_file = lock_file

        # Do a change in A
        clientb = TestClient(cache_folder=client.cache_folder)
        clientb.run("config set general.default_package_id_mode=full_package_mode")
        clientb.save({"conanfile.py": conanfile.format(requires=''),
                     "myfile.txt": "ByeA World!!"})
        clientb.run("create . PkgA/0.2@user/channel")

        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        client.run("lock build-order conan.lock --json=build_order.json")
        master_lockfile = client.load("conan.lock")

        json_file = os.path.join(client.current_folder, "build_order.json")
        to_build = json.loads(load(json_file))
        lock_fileaux = master_lockfile
        while to_build:
            ref, _, _, _ = to_build[0].pop(0)
            client_aux = TestClient(cache_folder=client.cache_folder)
            client_aux.run("config set general.default_package_id_mode=full_package_mode")
            client_aux.save({LOCKFILE: lock_fileaux})
            client_aux.run("install %s --build=%s --lockfile=conan.lock "
                           "--lockfile-out=conan.lock" % (ref, ref))
            lock_fileaux = load(os.path.join(client_aux.current_folder, "conan.lock"))
            client.save({"new_lock/conan.lock": lock_fileaux})
            client.run("lock update conan.lock new_lock/conan.lock")
            client.run("lock build-order conan.lock")
            lock_fileaux = client.load("conan.lock")
            output = str(client.out).splitlines()[-1]
            to_build = eval(output)

        new_lockfile = client.load(LOCKFILE)
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgB/0.1@user/channel: DEP FILE PkgA: ByeA World!!", client.out)
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgA: ByeA World!!", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgA: ByeA World!!", client.out)

        client.save({LOCKFILE: initial_lock_file})
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgB/0.1@user/channel: DEP FILE PkgA: HelloA", client.out)
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgA: HelloA", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgA: HelloA", client.out)

        client.save({LOCKFILE: new_lockfile})
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgB/0.1@user/channel: DEP FILE PkgA: ByeA World!!", client.out)
        self.assertIn("PkgC/0.1@user/channel: DEP FILE PkgA: ByeA World!!", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgA: ByeA World!!", client.out)

    def test_override(self):
        client = TestClient()
        client.run("config set general.default_package_id_mode=full_package_mode")

        # The original (unresolved) graph shows the requirements how they are
        # specified in the package recipes.  The graph contains conflicts:
        # There are three different versions of PkgA and two different versions
        # of PkgB.
        #
        # The overridden (resolved) graph shows the requirements after
        # conflicts have been resolved by overriding required package versions
        # in package PkgE.  This graph contains only one version per package.
        #
        # Original (unresolved) graph:       :  Overridden (resolved) graph:
        # ============================       :  ============================
        #                                    :
        # PkgA/0.1   PkgA/0.2    PkgA/0.3    :                       PkgA/0.3
        #    |          |           |\____   :       __________________/|
        #    |          |           |     \  :      /                   |
        # PkgB/0.1      |        PkgB/0.3 |  :  PkgB/0.1                |
        #    |\____     | _________/      |  :     |\_______   ________/|
        #    |     \    |/                |  :     |        \ /         |
        #    |     | PkgC/0.2             |  :     |      PkgC/0.2      |
        #    |     |    |                 |  :     |         |          |
        #    |     |    |                 |  :     |         |          |
        # PkgD/0.1 |    |                 |  :  PkgD/0.1     |          |
        #    | ___/____/_________________/   :     | _______/__________/
        #    |/                              :     |/
        # PkgE/0.1                           :  PkgE/0.1

        # PkgA/0.1
        client.save({
            "conanfile.py": conanfile.format(requires=""),
            "myfile.txt": "This is PkgA/0.1!",
        })
        client.run("export . PkgA/0.1@user/channel")

        # PkgA/0.2
        client.save({
            "conanfile.py": conanfile.format(requires=""),
            "myfile.txt": "This is PkgA/0.2!",
        })
        client.run("export . PkgA/0.2@user/channel")

        # PkgA/0.3
        client.save({
            "conanfile.py": conanfile.format(requires=""),
            "myfile.txt": "This is PkgA/0.3!",
        })
        client.run("export . PkgA/0.3@user/channel")

        # PkgB/0.1
        client.save({
            "conanfile.py": conanfile.format(requires='requires="PkgA/0.1@user/channel"'),
            "myfile.txt": "This is PkgB/0.1!",
        })
        client.run("export . PkgB/0.1@user/channel")

        # PkgB/0.3
        client.save({
            "conanfile.py": conanfile.format(requires='requires="PkgA/0.3@user/channel"'),
            "myfile.txt": "This is PkgB/0.3!",
        })
        client.run("export . PkgB/0.3@user/channel")

        # PkgC/0.2
        client.save({
            "conanfile.py": conanfile.format(requires=textwrap.dedent("""
                # This comment line is required to yield the correct indentation
                    def requirements(self):
                        self.requires("PkgA/0.2@user/channel", override=True)
                        self.requires("PkgB/0.3@user/channel", override=False)
                """)),
            "myfile.txt": "This is PkgC/0.2!",
        })
        client.run("export . PkgC/0.2@user/channel")

        # PkgD/0.1
        client.save({
            "conanfile.py": conanfile.format(requires='requires="PkgB/0.1@user/channel"'),
            "myfile.txt": "This is PkgD/0.1!",
        })
        client.run("export . PkgD/0.1@user/channel")

        # PkgE/0.1
        client.save({
            "conanfile.py": conanfile.format(requires=textwrap.dedent("""
                # This comment line is required to yield the correct indentation
                    def requirements(self):
                        self.requires("PkgA/0.3@user/channel", override=True)
                        self.requires("PkgB/0.1@user/channel", override=True)
                        self.requires("PkgC/0.2@user/channel", override=False)
                        self.requires("PkgD/0.1@user/channel", override=False)
                """)),
            "myfile.txt": "This is PkgE/0.1!",
        })
        client.run("export . PkgE/0.1@user/channel")

        client.run("lock create --reference=PkgE/0.1@user/channel --lockfile-out=master.lock")

        while True:
            client.run("lock build-order master.lock --json=build_order.json")
            json_file = os.path.join(client.current_folder, "build_order.json")
            to_build = json.loads(load(json_file))
            if not to_build:
                break
            ref, _, _, _ = to_build[0].pop(0)
            client.run("lock create --reference=%s --lockfile=master.lock "
                       "--lockfile-out=derived.lock" % ref)
            client.run("install %s --build=%s --lockfile=derived.lock "
                       "--lockfile-out=update.lock" % (ref, ref))
            client.run("lock update master.lock update.lock")

        client.run("install PkgE/0.1@user/channel --lockfile=master.lock")
        filtered_output = [
            line for line in str(client.out).splitlines()
            if any(pattern in line for pattern in ["SELF FILE", "DEP FILE"])
        ]
        expected_output = [
            "PkgA/0.3@user/channel: SELF FILE: This is PkgA/0.3!",
            "PkgB/0.1@user/channel: DEP FILE PkgA: This is PkgA/0.3!",
            "PkgB/0.1@user/channel: SELF FILE: This is PkgB/0.1!",
            "PkgC/0.2@user/channel: DEP FILE PkgA: This is PkgA/0.3!",
            "PkgC/0.2@user/channel: DEP FILE PkgB: This is PkgB/0.1!",
            "PkgC/0.2@user/channel: SELF FILE: This is PkgC/0.2!",
            "PkgD/0.1@user/channel: DEP FILE PkgA: This is PkgA/0.3!",
            "PkgD/0.1@user/channel: DEP FILE PkgB: This is PkgB/0.1!",
            "PkgD/0.1@user/channel: SELF FILE: This is PkgD/0.1!",
            "PkgE/0.1@user/channel: DEP FILE PkgA: This is PkgA/0.3!",
            "PkgE/0.1@user/channel: DEP FILE PkgB: This is PkgB/0.1!",
            "PkgE/0.1@user/channel: DEP FILE PkgC: This is PkgC/0.2!",
            "PkgE/0.1@user/channel: DEP FILE PkgD: This is PkgD/0.1!",
            "PkgE/0.1@user/channel: SELF FILE: This is PkgE/0.1!",
        ]
        self.assertListEqual(
            sorted(filtered_output),
            sorted(expected_output),
            msg = "Original client output:\n%s" % client.out,
        )

    def test_options(self):
        conanfile = textwrap.dedent("""
            from conans import ConanFile
            class Pkg(ConanFile):
                {requires}
                options = {{"myoption": [1, 2, 3, 4, 5]}}
                default_options = {{"myoption": 1}}
                def build(self):
                    self.output.info("BUILDING WITH OPTION: %s!!" % self.options.myoption)
                def package_info(self):
                    self.output.info("PACKAGE_INFO OPTION: %s!!" % self.options.myoption)
                """)
        client = TestClient()
        client.save({"conanfile.py": conanfile.format(requires="")})
        client.run("export . PkgA/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(requires='requires="PkgA/0.1@user/channel"')})
        client.run("export . PkgB/0.1@user/channel")
        client.save({"conanfile.py": conanfile.format(requires='requires="PkgB/0.1@user/channel"')})
        client.run("export . PkgC/0.1@user/channel")
        conanfiled = conanfile.format(requires='requires="PkgC/0.1@user/channel"')
        conanfiled = conanfiled.replace('default_options = {"myoption": 1}',
                                        'default_options = {"myoption": 2, "PkgC:myoption": 3,'
                                        '"PkgB:myoption": 4, "PkgA:myoption": 5}')
        client.save({"conanfile.py": conanfiled})
        client.run("export . PkgD/0.1@user/channel")

        client.run("profile new myprofile")
        # To make sure we can provide a profile as input
        client.run("lock create --reference=PkgD/0.1@user/channel -pr=myprofile "
                   "--lockfile-out=conan.lock")
        lock_file = client.load(LOCKFILE)

        client2 = TestClient(cache_folder=client.cache_folder)
        client2.save({"conanfile.py": conanfile.format(requires=""), LOCKFILE: lock_file})
        client2.run("create . PkgA/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgA/0.1@user/channel: BUILDING WITH OPTION: 5!!", client2.out)
        self.assertIn("PkgA/0.1@user/channel: PACKAGE_INFO OPTION: 5!!", client2.out)

        client2.save({"conanfile.py": conanfile.format(
            requires='requires="PkgA/0.1@user/channel"')})
        client2.run("create . PkgB/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgB/0.1@user/channel: PACKAGE_INFO OPTION: 4!!", client2.out)
        self.assertIn("PkgB/0.1@user/channel: BUILDING WITH OPTION: 4!!", client2.out)
        self.assertIn("PkgA/0.1@user/channel: PACKAGE_INFO OPTION: 5!!", client2.out)

        client2.save({"conanfile.py": conanfile.format(
            requires='requires="PkgB/0.1@user/channel"')})
        client2.run("create . PkgC/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgC/0.1@user/channel: PACKAGE_INFO OPTION: 3!!", client2.out)
        self.assertIn("PkgC/0.1@user/channel: BUILDING WITH OPTION: 3!!", client2.out)
        self.assertIn("PkgB/0.1@user/channel: PACKAGE_INFO OPTION: 4!!", client2.out)
        self.assertIn("PkgA/0.1@user/channel: PACKAGE_INFO OPTION: 5!!", client2.out)

        client2.save({"conanfile.py": conanfiled})
        client2.run("create . PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgD/0.1@user/channel: PACKAGE_INFO OPTION: 2!!", client2.out)
        self.assertIn("PkgD/0.1@user/channel: BUILDING WITH OPTION: 2!!", client2.out)
        self.assertIn("PkgC/0.1@user/channel: PACKAGE_INFO OPTION: 3!!", client2.out)
        self.assertIn("PkgB/0.1@user/channel: PACKAGE_INFO OPTION: 4!!", client2.out)
        self.assertIn("PkgA/0.1@user/channel: PACKAGE_INFO OPTION: 5!!", client2.out)

    @pytest.mark.skipif(not get_env("TESTING_REVISIONS_ENABLED", False), reason="Only revisions")
    def test_package_revisions_unkown_id_update(self):
        # https://github.com/conan-io/conan/issues/7588
        client = TestClient()
        client.run("config set general.default_package_id_mode=package_revision_mode")
        files = {
            "pkga/conanfile.py": conanfile.format(requires=""),
            "pkga/myfile.txt": "HelloA",
            "pkgb/conanfile.py": conanfile.format(requires='requires="PkgA/[*]@user/channel"'),
            "pkgb/myfile.txt": "HelloB",
            "pkgc/conanfile.py": conanfile.format(requires='requires="PkgB/[*]@user/channel"'),
            "pkgc/myfile.txt": "HelloC",
            "pkgd/conanfile.py": conanfile.format(requires='requires="PkgC/[*]@user/channel"'),
            "pkgd/myfile.txt": "HelloD",
        }
        client.save(files)
        client.run("export pkga PkgA/0.1@user/channel")
        client.run("export pkgb PkgB/0.1@user/channel")
        client.run("export pkgc PkgC/0.1@user/channel")
        client.run("export pkgd PkgD/0.1@user/channel")

        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        lockfile = json.loads(client.load("conan.lock"))
        nodes = lockfile["graph_lock"]["nodes"]
        self.assertEqual(nodes["3"]["ref"], "PkgB/0.1@user/channel#fa97c46bf83849a5db4564327b3cfada")
        self.assertEqual(nodes["3"]["package_id"], "Package_ID_unknown")

        client.run("install PkgA/0.1@user/channel --build=PkgA/0.1@user/channel "
                   "--lockfile=conan.lock --lockfile-out=conan_out.lock")
        client.run("lock update conan.lock conan_out.lock")

        client.run("install PkgB/0.1@user/channel --build=PkgB/0.1@user/channel "
                   "--lockfile=conan.lock --lockfile-out=conan_out.lock")
        lockfile = json.loads(client.load("conan_out.lock"))
        nodes = lockfile["graph_lock"]["nodes"]
        self.assertEqual(nodes["3"]["ref"], "PkgB/0.1@user/channel#fa97c46bf83849a5db4564327b3cfada")
        self.assertEqual(nodes["3"]["package_id"], "6e9742c2106791c1c777da8ccfb12a1408385d8d")
        self.assertEqual(nodes["3"]["prev"], "f971905c142e0de728f32a7237553622")

        client.run("lock update conan.lock conan_out.lock")
        lockfile = json.loads(client.load("conan.lock"))
        nodes = lockfile["graph_lock"]["nodes"]
        self.assertEqual(nodes["3"]["ref"], "PkgB/0.1@user/channel#fa97c46bf83849a5db4564327b3cfada")
        self.assertEqual(nodes["3"]["package_id"], "6e9742c2106791c1c777da8ccfb12a1408385d8d")
        self.assertEqual(nodes["3"]["prev"], "f971905c142e0de728f32a7237553622")


class CIPythonRequiresTest(unittest.TestCase):
    python_req = textwrap.dedent("""
        from conans import ConanFile
        def msg(conanfile):
            conanfile.output.info("{}")
        class Pkg(ConanFile):
            pass
        """)

    consumer = textwrap.dedent("""
        from conans import ConanFile, load
        import os
        class Pkg(ConanFile):
            {requires}
            python_requires = "pyreq/[*]@user/channel"
            def package_info(self):
                self.python_requires["pyreq"].module.msg(self)
            """)

    def setUp(self):
        client = TestClient()
        client.run("config set general.default_package_id_mode=full_package_mode")
        client.save({"conanfile.py": self.python_req.format("HelloPyWorld")})
        client.run("export . pyreq/0.1@user/channel")

        client.save({"conanfile.py": self.consumer.format(requires="")})
        client.run("create . PkgA/0.1@user/channel")
        client.save(
            {"conanfile.py": self.consumer.format(requires='requires="PkgA/0.1@user/channel"')})
        client.run("create . PkgB/0.1@user/channel")
        client.save(
            {"conanfile.py": self.consumer.format(requires='requires="PkgB/[~0]@user/channel"')})
        client.run("create . PkgC/0.1@user/channel")
        client.save(
            {"conanfile.py": self.consumer.format(requires='requires="PkgC/0.1@user/channel"')})
        client.run("create . PkgD/0.1@user/channel")
        for pkg in ("PkgA", "PkgB", "PkgC", "PkgD"):
            self.assertIn("{}/0.1@user/channel: HelloPyWorld".format(pkg), client.out)

        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        self.client = client

    def test_version_ranges(self):
        client = self.client
        initial_lockfile = client.load("conan.lock")
        # Do a change in python_require
        client.save({"conanfile.py": self.python_req.format("ByePyWorld")})
        client.run("export . pyreq/0.2@user/channel")

        # Go back to main orchestrator
        client.run("lock create --reference=PkgD/0.1@user/channel --lockfile-out=conan.lock")
        client.run("lock build-order conan.lock --json=build_order.json")
        master_lockfile = client.load("conan.lock")
        json_file = client.load("build_order.json")
        to_build = json.loads(json_file)
        lock_fileaux = master_lockfile
        while to_build:
            for ref, _, _, _ in to_build[0]:
                client_aux = TestClient(cache_folder=client.cache_folder)
                client_aux.save({"conan.lock": lock_fileaux})
                client_aux.run("install %s --build=%s --lockfile=conan.lock "
                               "--lockfile-out=conan.lock" % (ref, ref))
                lock_fileaux = client_aux.load("conan.lock")
                client.save({"new_lock/conan.lock": lock_fileaux})
                client.run("lock update conan.lock new_lock/conan.lock")

            client.run("lock build-order conan.lock --json=bo.json")
            lock_fileaux = client.load("conan.lock")
            to_build = json.loads(client.load("bo.json"))

        new_lockfile = client.load("conan.lock")
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        for pkg in ("PkgA", "PkgB", "PkgC", "PkgD"):
            self.assertIn("{}/0.1@user/channel: ByePyWorld".format(pkg), client.out)

        client.save({"conan.lock": initial_lockfile})
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        for pkg in ("PkgA", "PkgB", "PkgC", "PkgD"):
            self.assertIn("{}/0.1@user/channel: HelloPyWorld".format(pkg), client.out)

        client.save({"conan.lock": new_lockfile})
        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        for pkg in ("PkgA", "PkgB", "PkgC", "PkgD"):
            self.assertIn("{}/0.1@user/channel: ByePyWorld".format(pkg), client.out)

    def test_version_ranges_partial_unused(self):
        client = self.client
        consumer = self.consumer
        # Do a change in B
        client.save({"conanfile.py": consumer.format(requires='requires="PkgA/0.1@user/channel"')})
        client.run("lock create conanfile.py --name=PkgB --version=1.0 --user=user "
                   "--channel=channel --lockfile-out=buildb.lock")

        # Do a change in python_require
        client.save({"conanfile.py": self.python_req.format("ByePyWorld")})
        client.run("export . pyreq/0.2@user/channel")

        # create the package with the previous version of python_require
        client.save({"conanfile.py": consumer.format(requires='requires="PkgA/0.1@user/channel"')})
        # It is a new version, it will not be used in the product build!
        client.run("create . PkgB/1.0@user/channel --lockfile=buildb.lock")
        self.assertIn("pyreq/0.1", client.out)
        self.assertNotIn("pyreq/0.2", client.out)

        # Go back to main orchestrator
        # This should fail, as PkgB/1.0 is not involved in the new resolution
        client.run("lock create --reference=PkgD/0.1@user/channel "
                   "--lockfile=buildb.lock --lockfile-out=error.lock")
        # User can perfectly go and check the resulting lockfile and check if PkgB/0.1 is there
        # We can probably help automate this with a "conan lock find" subcommand
        error_lock = client.load("error.lock")
        self.assertNotIn("PkgB/1.0@user/channel", error_lock)

        client.run("lock build-order conan.lock --json=build_order.json")
        json_file = client.load("build_order.json")
        to_build = json.loads(json_file)
        self.assertEqual(to_build, [])

        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        for pkg in ("PkgA", "PkgB", "PkgC", "PkgD"):
            self.assertIn("{}/0.1@user/channel: HelloPyWorld".format(pkg), client.out)

        client.run("install PkgD/0.1@user/channel", assert_error=True)
        self.assertIn("ERROR: Missing prebuilt package", client.out)

        client.run("install PkgD/0.1@user/channel --build=missing")
        for pkg in ("PkgA", "PkgB", "PkgC", "PkgD"):
            self.assertIn("{}/0.1@user/channel: ByePyWorld".format(pkg), client.out)

    def test_version_ranges_partial(self):
        client = self.client
        consumer = self.consumer
        # Do a change in B
        client.save({"conanfile.py": consumer.format(requires='requires="PkgA/0.1@user/channel"')})
        client.run("lock create conanfile.py --name=PkgB --version=0.2 --user=user "
                   "--channel=channel --lockfile-out=buildb.lock")

        # Do a change in python_require
        client.save({"conanfile.py": self.python_req.format("ByePyWorld")})
        client.run("export . pyreq/0.2@user/channel")

        # create the package with the previous version of python_require
        client.save({"conanfile.py": consumer.format(requires='requires="PkgA/0.1@user/channel"')})
        # It is a new version, it will not be used in the product build!
        client.run("create . PkgB/0.2@user/channel --lockfile=buildb.lock")
        self.assertIn("pyreq/0.1", client.out)
        self.assertNotIn("pyreq/0.2", client.out)

        # Go back to main orchestrator
        client.run("lock create --reference=PkgD/0.1@user/channel "
                   "--lockfile=buildb.lock --lockfile-out=conan.lock")

        client.run("lock build-order conan.lock --json=build_order.json")
        json_file = client.load("build_order.json")
        to_build = json.loads(json_file)
        if client.cache.config.revisions_enabled:
            build_order = [[['PkgC/0.1@user/channel#9e5471ca39a16a120b25ee5690539c71',
                             'bca7337f8d2fde6cdc9dd17cdc56bc0b0a0e352d', 'host', '4']],
                           [['PkgD/0.1@user/channel#068fd3ce2a88181dff0b44de344a93a4',
                             '63a3463d4dd4cc8d7bca7a9fe5140abe582f349a', 'host', '3']]]
        else:
            build_order = [[['PkgC/0.1@user/channel',
                             'bca7337f8d2fde6cdc9dd17cdc56bc0b0a0e352d', 'host', '4']],
                           [['PkgD/0.1@user/channel',
                             '63a3463d4dd4cc8d7bca7a9fe5140abe582f349a', 'host', '3']]]
        self.assertEqual(to_build, build_order)

        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock --build=missing")
        self.assertIn("PkgA/0.1@user/channel: HelloPyWorld", client.out)
        self.assertIn("PkgB/0.2@user/channel: HelloPyWorld", client.out)
        self.assertIn("PkgC/0.1@user/channel: ByePyWorld", client.out)
        self.assertIn("PkgD/0.1@user/channel: ByePyWorld", client.out)

        client.run("install PkgD/0.1@user/channel", assert_error=True)
        self.assertIn("ERROR: Missing prebuilt package", client.out)

        client.run("install PkgD/0.1@user/channel --build=missing")
        self.assertIn("PkgA/0.1@user/channel: ByePyWorld", client.out)
        self.assertIn("PkgB/0.2@user/channel: ByePyWorld", client.out)
        self.assertIn("PkgC/0.1@user/channel: ByePyWorld", client.out)
        self.assertIn("PkgD/0.1@user/channel: ByePyWorld", client.out)

        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        self.assertIn("PkgA/0.1@user/channel: HelloPyWorld", client.out)
        self.assertIn("PkgB/0.2@user/channel: HelloPyWorld", client.out)
        self.assertIn("PkgC/0.1@user/channel: ByePyWorld", client.out)
        self.assertIn("PkgD/0.1@user/channel: ByePyWorld", client.out)


class CIBuildRequiresTest(unittest.TestCase):
    def test_version_ranges(self):
        client = TestClient()
        client.run("config set general.default_package_id_mode=full_package_mode")
        myprofile = textwrap.dedent("""
            [build_requires]
            br/[>=0.1]@user/channel
            """)
        files = {
            "myprofile": myprofile,
            "br/conanfile.py": GenConanfile(),
            "pkga/conanfile.py": conanfile.format(requires=""),
            "pkga/myfile.txt": "HelloA",
            "pkgb/conanfile.py": conanfile.format(requires='requires="PkgA/[*]@user/channel"'),
            "pkgb/myfile.txt": "HelloB",
            "pkgc/conanfile.py": conanfile.format(requires='requires="PkgB/[*]@user/channel"'),
            "pkgc/myfile.txt": "HelloC",
            "pkgd/conanfile.py": conanfile.format(requires='requires="PkgC/[*]@user/channel"'),
            "pkgd/myfile.txt": "HelloD",
        }
        client.save(files)
        client.run("create br br/0.1@user/channel")
        client.run("create pkga PkgA/0.1@user/channel -pr=myprofile")
        client.run("create pkgb PkgB/0.1@user/channel -pr=myprofile")
        client.run("create pkgc PkgC/0.1@user/channel -pr=myprofile")
        client.run("create pkgd PkgD/0.1@user/channel -pr=myprofile")

        self.assertIn("PkgD/0.1@user/channel: SELF FILE: HelloD", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgA: HelloA", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgB: HelloB", client.out)
        self.assertIn("PkgD/0.1@user/channel: DEP FILE PkgC: HelloC", client.out)

        # Go back to main orchestrator
        client.run("lock create --reference=PkgD/0.1@user/channel --build -pr=myprofile "
                   " --lockfile-out=conan.lock")

        # Do a change in br
        client.run("create br br/0.2@user/channel")

        client.run("lock build-order conan.lock --json=build_order.json")
        self.assertIn("br/0.1", client.out)
        self.assertNotIn("br/0.2", client.out)
        master_lockfile = client.load("conan.lock")

        json_file = client.load("build_order.json")
        to_build = json.loads(json_file)
        if client.cache.config.revisions_enabled:
            build_order = [[['br/0.1@user/channel#f3367e0e7d170aa12abccb175fee5f97',
                             '5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9', 'host', '5']],
                           [['PkgA/0.1@user/channel#189390ce059842ce984e0502c52cf736',
                             '5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9', 'host', '4']],
                           [['PkgB/0.1@user/channel#fa97c46bf83849a5db4564327b3cfada',
                             '096f747d204735584fa0115bcbd7482d424094bc', 'host', '3']],
                           [['PkgC/0.1@user/channel#c6f95948619d28d9d96b0ae86c46a482',
                             'f6d5dbb6f309dbf8519278bae8d07d3b739b3dec', 'host', '2']],
                           [['PkgD/0.1@user/channel#fce78c934bc0de73eeb05eb4060fc2b7',
                             'de4467a3fa6ef01b09b7464e85553fb4be2d2096', 'host', '1']]]
        else:
            build_order = [[['br/0.1@user/channel',
                             '5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9', 'host', '5']],
                           [['PkgA/0.1@user/channel',
                             '5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9', 'host', '4']],
                           [['PkgB/0.1@user/channel',
                             '096f747d204735584fa0115bcbd7482d424094bc', 'host', '3']],
                           [['PkgC/0.1@user/channel',
                             'f6d5dbb6f309dbf8519278bae8d07d3b739b3dec', 'host', '2']],
                           [['PkgD/0.1@user/channel',
                             'de4467a3fa6ef01b09b7464e85553fb4be2d2096', 'host', '1']]]

        self.assertEqual(to_build, build_order)
        lock_fileaux = master_lockfile
        while to_build:
            for ref, _, _, _ in to_build[0]:
                client_aux = TestClient(cache_folder=client.cache_folder)
                client_aux.save({LOCKFILE: lock_fileaux})
                client_aux.run("install %s --build=%s --lockfile=conan.lock "
                               "--lockfile-out=conan.lock" % (ref, ref))
                self.assertIn("br/0.1", client_aux.out)
                self.assertNotIn("br/0.2", client_aux.out)
                lock_fileaux = client_aux.load(LOCKFILE)
                client.save({"new_lock/%s" % LOCKFILE: lock_fileaux})
                client.run("lock update conan.lock new_lock/conan.lock")

            client.run("lock build-order conan.lock --json=build_order.json")
            lock_fileaux = client.load(LOCKFILE)
            to_build = json.loads(client.load("build_order.json"))

        client.run("install PkgD/0.1@user/channel --lockfile=conan.lock")
        # No build require at all
        self.assertNotIn("br/0.", client.out)

        client.run("install PkgD/0.1@user/channel --build -pr=myprofile")
        self.assertIn("br/0.2", client.out)
        self.assertNotIn("br/0.1", client.out)


class CIBuildRequiresTwoProfilesTest(unittest.TestCase):
    def test_version_ranges(self):
        client = TestClient()
        client.run("config set general.default_package_id_mode=full_package_mode")
        myprofile_host = textwrap.dedent("""
            [settings]
            os=Linux
            [build_requires]
            br/[>=0.1]
            """)
        myprofile_build = textwrap.dedent("""
           [settings]
           os=Windows
           """)
        conanfile_os = textwrap.dedent("""
            from conans import ConanFile, load
            from conans.tools import save
            import os
            class Pkg(ConanFile):
                settings = "os"
                {requires}

                keep_imports = True
                def imports(self):
                    self.copy("myfile.txt", folder=True)
                def package(self):
                    save(os.path.join(self.package_folder, "myfile.txt"),
                         "%s %s" % (self.name, self.settings.os))
                    self.copy("*myfile.txt")
                def package_info(self):
                    self.output.info("MYOS=%s!!!" % self.settings.os)
                    self.output.info("SELF FILE: %s"
                        % load(os.path.join(self.package_folder, "myfile.txt")))
                    for d in os.listdir(self.package_folder):
                        p = os.path.join(self.package_folder, d, "myfile.txt")
                        if os.path.isfile(p):
                            self.output.info("DEP FILE %s: %s" % (d, load(p)))
                """)
        files = {
            "profile_host": myprofile_host,
            "profile_build": myprofile_build,
            "br/conanfile.py": conanfile_os.format(requires=""),
            "pkga/conanfile.py": conanfile_os.format(requires=""),
            "pkgb/conanfile.py": conanfile_os.format(requires='requires="PkgA/[*]"'),
            "pkgc/conanfile.py": conanfile_os.format(requires='requires="PkgB/[*]"'),
            "pkgd/conanfile.py": conanfile_os.format(requires='requires="PkgC/[*]"'),
        }
        client.save(files)
        # Note the creating of BR is in the BUILD profile
        client.run("create br br/0.1@ --build-require -pr:h=profile_host -pr:b=profile_build")
        assert "br/0.1: SELF FILE: br Linux" not in client.out
        client.run("create pkga PkgA/0.1@ -pr:h=profile_host -pr:b=profile_build")
        client.run("create pkgb PkgB/0.1@ -pr:h=profile_host -pr:b=profile_build")
        client.run("create pkgc PkgC/0.1@ -pr:h=profile_host -pr:b=profile_build")
        client.run("create pkgd PkgD/0.1@ -pr:h=profile_host -pr:b=profile_build")

        self.assertIn("PkgD/0.1: SELF FILE: PkgD Linux", client.out)
        self.assertIn("PkgD/0.1: DEP FILE PkgA: PkgA Linux", client.out)
        self.assertIn("PkgD/0.1: DEP FILE PkgB: PkgB Linux", client.out)
        self.assertIn("PkgD/0.1: DEP FILE PkgC: PkgC Linux", client.out)

        # Go back to main orchestrator
        client.run("lock create --reference=PkgD/0.1@ --build -pr:h=profile_host -pr:b=profile_build"
                   " --lockfile-out=conan.lock")

        # Do a change in br
        client.run("create br br/0.2@ ")

        client.run("lock build-order conan.lock --json=build_order.json")
        self.assertIn("br/0.1", client.out)
        self.assertNotIn("br/0.2", client.out)
        master_lockfile = client.load("conan.lock")

        json_file = client.load("build_order.json")
        to_build = json.loads(json_file)
        if client.cache.config.revisions_enabled:
            build_order = [[['br/0.1@#583b8302673adce66f12f2bec01fe9c3',
                             '3475bd55b91ae904ac96fde0f106a136ab951a5e', 'build', '5']],
                           [['PkgA/0.1@#583b8302673adce66f12f2bec01fe9c3',
                             'cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31', 'host', '4']],
                           [['PkgB/0.1@#4b1da86739946fe16a9545d1f6bc9022',
                             '4a87f1e30266a1c1c685c0904cfb137a3dba11c7', 'host', '3']],
                           [['PkgC/0.1@#3e1048668b2a795f6742d04971f11a7d',
                             '50ad117314ca51a58e427a26f264e27e79b94cd4', 'host', '2']],
                           [['PkgD/0.1@#e6cc0ca095ca32bba1a6dff0af6f4eb3',
                             'e66cc39a683367fdd17218bdbab7d6e95c0414e1', 'host', '1']]]
        else:
            build_order = [[['br/0.1@', '3475bd55b91ae904ac96fde0f106a136ab951a5e', 'build', '5']],
                           [['PkgA/0.1@', 'cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31', 'host', '4']],
                           [['PkgB/0.1@', '4a87f1e30266a1c1c685c0904cfb137a3dba11c7', 'host', '3']],
                           [['PkgC/0.1@', '50ad117314ca51a58e427a26f264e27e79b94cd4', 'host', '2']],
                           [['PkgD/0.1@', 'e66cc39a683367fdd17218bdbab7d6e95c0414e1', 'host', '1']]]

        self.assertEqual(to_build, build_order)
        lock_fileaux = master_lockfile
        while to_build:
            for ref, _, build, _ in to_build[0]:
                client_aux = TestClient(cache_folder=client.cache_folder)
                client_aux.save({LOCKFILE: lock_fileaux})
                is_build_require = "--build-require" if build == "build" else ""
                client_aux.run("install %s --build=%s --lockfile=conan.lock "
                               "--lockfile-out=conan.lock %s" % (ref, ref, is_build_require))
                assert "br/0.1: SELF FILE: br Windows" in client_aux.out
                self.assertIn("br/0.1", client_aux.out)
                self.assertNotIn("br/0.2", client_aux.out)
                lock_fileaux = client_aux.load(LOCKFILE)
                client.save({"new_lock/%s" % LOCKFILE: lock_fileaux})
                client.run("lock update conan.lock new_lock/conan.lock")

            client.run("lock build-order conan.lock --json=build_order.json")
            lock_fileaux = client.load(LOCKFILE)
            to_build = json.loads(client.load("build_order.json"))

        client.run("install PkgD/0.1@ --lockfile=conan.lock")
        # No build require at all
        self.assertNotIn("br/0.", client.out)

        client.run("install PkgD/0.1@ --build  -pr:h=profile_host -pr:b=profile_build")
        self.assertIn("br/0.2", client.out)
        self.assertNotIn("br/0.1", client.out)


class CIPrivateRequiresTest(unittest.TestCase):
    def test(self):
        # https://github.com/conan-io/conan/issues/7985
        client = TestClient()
        files = {
            "private/conanfile.py": GenConanfile().with_option("myoption", [True, False]),
            "pkga/conanfile.py": textwrap.dedent("""
                from conans import ConanFile
                class PkgA(ConanFile):
                    requires = ("private/0.1", "private"),
                    def configure(self):
                        self.options["private"].myoption = True
                """),
            "pkgb/conanfile.py": textwrap.dedent("""
                from conans import ConanFile
                class PkgB(ConanFile):
                    requires = "pkga/0.1", ("private/0.1", "private"),
                    def configure(self):
                        self.options["private"].myoption = False
                """),
            "pkgc/conanfile.py": GenConanfile().with_require("pkgb/0.1")
        }
        client.save(files)
        client.run("export private private/0.1@")
        client.run("export pkga pkga/0.1@")
        client.run("export pkgb pkgb/0.1@")

        client.run("lock create pkgc/conanfile.py --name=pkgc --version=0.1 --build "
                   "--lockfile-out=conan.lock")
        client.run("lock build-order conan.lock --json=build_order.json")
        json_file = client.load("build_order.json")
        to_build = json.loads(json_file)
        if client.cache.config.revisions_enabled:
            build_order = [[['private/0.1@#e31c7a656abb86256b08af0e64d37d42',
                             'd2560ba1787c188a1d7fabeb5f8e012ac53301bb', 'host', '3'],
                            ['private/0.1@#e31c7a656abb86256b08af0e64d37d42',
                             '5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9', 'host', '4']],
                           [['pkga/0.1@#edf085c091f9f4adfdb623bda9415a79',
                             '5b0fc4382d9c849ae3ef02a57b62b26ad5137990', 'host', '2']],
                           [['pkgb/0.1@#9b0edf8f61a88f92e05919b406d74089',
                             'd7d6ac48b43e368b0a5ff79015acea49b758ffdf', 'host', '1']]]
        else:
            build_order = [[['private/0.1@',
                             'd2560ba1787c188a1d7fabeb5f8e012ac53301bb', 'host', '3'],
                            ['private/0.1@',
                             '5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9', 'host', '4']],
                           [['pkga/0.1@',
                             '5b0fc4382d9c849ae3ef02a57b62b26ad5137990', 'host', '2']],
                           [['pkgb/0.1@',
                             'd7d6ac48b43e368b0a5ff79015acea49b758ffdf', 'host', '1']]]

        self.assertEqual(to_build, build_order)

        for ref, pid, _, node_id in build_order[0]:
            client.run("install %s --build=%s --lockfile=conan.lock --lockfile-out=conan.lock "
                       "--lockfile-node-id=%s" % (ref, ref, node_id))
            self.assertIn('private/0.1:{} - Build'.format(pid), client.out)



import os

import pytest
import textwrap

from conans.test.assets.sources import gen_function_cpp
from conans.test.functional.toolchains.meson._base import TestMesonBase


@pytest.mark.tool_pkg_config
class MesonTest(TestMesonBase):
    _test_package_meson_build = textwrap.dedent("""
        project('test_package', 'cpp')
        hello = dependency('hello', version : '>=0.1')
        test_package = executable('test_package', 'test_package.cpp', dependencies: hello)
        test('test package', test_package)
        """)

    _test_package_conanfile_py = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.meson import Meson, MesonToolchain


        class TestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "pkg_config"

            def layout(self):
                self.folders.build = "build"

            def generate(self):
                tc = MesonToolchain(self)
                tc.generate()

            def build(self):
                meson = Meson(self)
                meson.configure()
                meson.build()

            def test(self):
                meson = Meson(self)
                meson.configure()
                meson.test()
        """)

    def test_reuse(self):
        self.t.run("new hello/0.1 -s")

        test_package_cpp = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])

        self.t.save({os.path.join("test_package", "conanfile.py"): self._test_package_conanfile_py,
                     os.path.join("test_package", "meson.build"): self._test_package_meson_build,
                     os.path.join("test_package", "test_package.cpp"): test_package_cpp})

        self.t.run("create . hello/0.1@ %s" % self._settings_str)

        self._check_binary()



import glob
import os
import platform
import textwrap
import unittest

import pytest

from conans.test.assets.genconanfile import GenConanfile
from conans.test.utils.tools import TestClient
from conans.util.files import load


def get_requires_from_content(content):
    for line in content.splitlines():
        if "Requires:" in line:
            return line
    return ""


class PkgGeneratorTest(unittest.TestCase):

    # Without compiler, def rpath_flags(settings, os_build, lib_paths): doesn't append the -Wl...etc
    def test_pkg_config_dirs(self):
        # https://github.com/conan-io/conan/issues/2756
        conanfile = """
import os
from conans import ConanFile

class PkgConfigConan(ConanFile):
    name = "MyLib"
    version = "0.1"

    def package_info(self):
        self.cpp_info.frameworkdirs = []
        self.cpp_info.filter_empty = False
        libname = "mylib"
        fake_dir = os.path.join("/", "my_absoulte_path", "fake")
        include_dir = os.path.join(fake_dir, libname, "include")
        lib_dir = os.path.join(fake_dir, libname, "lib")
        lib_dir2 = os.path.join(self.package_folder, "lib2")
        self.cpp_info.includedirs = [include_dir]
        self.cpp_info.libdirs = [lib_dir, lib_dir2]
"""
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . danimtb/testing")
        client.run("install MyLib/0.1@danimtb/testing -g pkg_config")

        pc_path = os.path.join(client.current_folder, "MyLib.pc")
        self.assertTrue(os.path.exists(pc_path))
        pc_content = load(pc_path)
        expected_rpaths = ""
        if platform.system() in ("Linux", "Darwin"):
            expected_rpaths = ' -Wl,-rpath,"${libdir}" -Wl,-rpath,"${libdir2}"'
        expected_content = """libdir=/my_absoulte_path/fake/mylib/lib
libdir2=${prefix}/lib2
includedir=/my_absoulte_path/fake/mylib/include

Name: MyLib
Description: Conan package: MyLib
Version: 0.1
Libs: -L"${libdir}" -L"${libdir2}"%s
Cflags: -I"${includedir}"\
""" % expected_rpaths
        self.assertEqual("\n".join(pc_content.splitlines()[1:]), expected_content)

        def assert_is_abs(path):
            self.assertTrue(os.path.isabs(path))

        for line in pc_content.splitlines():
            if line.startswith("includedir="):
                assert_is_abs(line[len("includedir="):])
                self.assertTrue(line.endswith("include"))
            elif line.startswith("libdir="):
                assert_is_abs(line[len("libdir="):])
                self.assertTrue(line.endswith("lib"))
            elif line.startswith("libdir3="):
                self.assertIn("${prefix}/lib2", line)

    def test_pkg_config_without_libdir(self):
        conanfile = """
import os
from conans import ConanFile

class PkgConfigConan(ConanFile):
    name = "MyLib"
    version = "0.1"

    def package_info(self):
        self.cpp_info.includedirs = []
        self.cpp_info.libdirs = []
        self.cpp_info.bindirs = []
        self.cpp_info.libs = []
"""
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . danimtb/testing")
        client.run("install MyLib/0.1@danimtb/testing -g pkg_config")

        pc_path = os.path.join(client.current_folder, "MyLib.pc")
        self.assertTrue(os.path.exists(pc_path))
        pc_content = load(pc_path)
        expected = textwrap.dedent("""
            Name: MyLib
            Description: Conan package: MyLib
            Version: 0.1
            Libs:%s
            Cflags: """ % " ")  # ugly hack for trailing whitespace removed by IDEs
        self.assertEqual("\n".join(pc_content.splitlines()[1:]), expected)

    def test_pkg_config_rpaths(self):
        # rpath flags are only generated for gcc and clang
        profile = """
[settings]
os=Linux
compiler=gcc
compiler.version=7
compiler.libcxx=libstdc++
"""
        conanfile = """
from conans import ConanFile

class PkgConfigConan(ConanFile):
    name = "MyLib"
    version = "0.1"
    settings = "os", "compiler"
    exports = "mylib.so"

    def package(self):
        self.copy("mylib.so", dst="lib")

    def package_info(self):
        self.cpp_info.libs = ["mylib"]
"""
        client = TestClient()
        client.save({"conanfile.py": conanfile,
                     "linux_gcc": profile,
                     "mylib.so": "fake lib content"})
        client.run("create . danimtb/testing -pr=linux_gcc")
        client.run("install MyLib/0.1@danimtb/testing -g pkg_config -pr=linux_gcc")

        pc_path = os.path.join(client.current_folder, "MyLib.pc")
        self.assertTrue(os.path.exists(pc_path))
        pc_content = load(pc_path)
        self.assertIn("-Wl,-rpath,\"${libdir}\"", pc_content)

    def test_system_libs(self):
        conanfile = """
from conans import ConanFile
from conans.tools import save
import os

class PkgConfigConan(ConanFile):
    name = "MyLib"
    version = "0.1"

    def package(self):
        save(os.path.join(self.package_folder, "lib", "file"), "")

    def package_info(self):
        self.cpp_info.libs = ["mylib1", "mylib2"]
        self.cpp_info.system_libs = ["system_lib1", "system_lib2"]
"""
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create .")
        client.run("install MyLib/0.1@ -g pkg_config")

        pc_content = client.load("MyLib.pc")
        self.assertIn('Libs: -L"${libdir}" -lmylib1  -lmylib2  -lsystem_lib1  -lsystem_lib2 ',
                      pc_content)

    def test_multiple_include(self):
        # https://github.com/conan-io/conan/issues/7056
        conanfile = textwrap.dedent("""
            from conans import ConanFile
            from conans.tools import save
            import os

            class PkgConfigConan(ConanFile):
                def package(self):
                    for p in ["inc1", "inc2", "inc3/foo", "lib1", "lib2"]:
                        save(os.path.join(self.package_folder, p, "file"), "")

                def package_info(self):
                    self.cpp_info.includedirs = ["inc1", "inc2", "inc3/foo"]
                    self.cpp_info.libdirs = ["lib1", "lib2"]
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . pkg/0.1@")
        client.run("install pkg/0.1@ -g pkg_config")

        pc_content = client.load("pkg.pc")
        self.assertIn("includedir=${prefix}/inc1", pc_content)
        self.assertIn("includedir2=${prefix}/inc2", pc_content)
        self.assertIn("includedir3=${prefix}/inc3/foo", pc_content)
        self.assertIn("libdir=${prefix}/lib1", pc_content)
        self.assertIn("libdir2=${prefix}/lib2", pc_content)
        self.assertIn('Libs: -L"${libdir}" -L"${libdir2}"', pc_content)
        self.assertIn('Cflags: -I"${includedir}" -I"${includedir2}" -I"${includedir3}"', pc_content)

    def test_empty_include(self):
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . pkg/0.1@")
        client.run("install pkg/0.1@ -g=pkg_config")
        pc = client.load("pkg.pc")
        self.assertNotIn("libdir=${prefix}/lib", pc)
        self.assertNotIn("includedir=${prefix}/include", pc)

    def test_custom_content(self):
        # https://github.com/conan-io/conan/issues/7661
        conanfile = textwrap.dedent("""
            from conans import ConanFile
            from conans.tools import save
            import os
            import textwrap

            class PkgConfigConan(ConanFile):
                def package(self):
                    save(os.path.join(self.package_folder, "include" ,"file"), "")
                    save(os.path.join(self.package_folder, "lib" ,"file"), "")

                def package_info(self):
                    custom_content = textwrap.dedent(\"""
                            datadir=${prefix}/share
                            schemasdir=${datadir}/mylib/schemas
                            bindir=${prefix}/bin
                        \""")
                    self.cpp_info.set_property("pkg_config_custom_content", custom_content)
                    self.cpp_info.includedirs = ["include"]
                    self.cpp_info.libdirs = ["lib"]
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . pkg/0.1@")
        client.run("install pkg/0.1@ -g pkg_config")

        pc_content = client.load("pkg.pc")
        self.assertIn("libdir=${prefix}/lib", pc_content)
        self.assertIn("datadir=${prefix}/share", pc_content)
        self.assertIn("schemasdir=${datadir}/mylib/schemas", pc_content)
        self.assertIn("bindir=${prefix}/bin", pc_content)
        self.assertIn("Name: pkg", pc_content)

    def test_custom_content_components(self):
        conanfile = textwrap.dedent("""
            from conans import ConanFile
            from conans.tools import save
            import os
            import textwrap

            class PkgConfigConan(ConanFile):
                def package_info(self):
                    self.cpp_info.components["mycomponent"].set_property("pkg_config_custom_content",
                                                                         "componentdir=${prefix}/mydir")
                    self.cpp_info.components["mycomponent"].set_property("component_version",
                                                                         "19.8.199")
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . pkg/0.1@")
        client.run("install pkg/0.1@ -g pkg_config")

        pc_content = client.load("mycomponent.pc")
        self.assertIn("componentdir=${prefix}/mydir", pc_content)
        self.assertIn("Version: 19.8.199", pc_content)


def test_components_and_package_pc_creation_order():
    """
    Testing if the root package PC file name matches with any of the components one, the first one
    is not going to be created. Components have more priority than root package.

    Issue related: https://github.com/conan-io/conan/issues/10341
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conans import ConanFile

        class PkgConfigConan(ConanFile):

            def package_info(self):
                self.cpp_info.set_property("pkg_config_name", "OpenCL")
                self.cpp_info.components["_opencl-headers"].set_property("pkg_config_name", "OpenCL")
                self.cpp_info.components["_opencl-other"].set_property("pkg_config_name", "OtherCL")
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . opencl/1.0@")

    conanfile = textwrap.dedent("""
        from conans import ConanFile

        class PkgConfigConan(ConanFile):
            requires = "opencl/1.0"

            def package_info(self):
                self.cpp_info.components["comp"].set_property("pkg_config_name", "pkgb")
                self.cpp_info.components["comp"].requires.append("opencl::_opencl-headers")
        """)
    client.save({"conanfile.py": conanfile}, clean_first=True)
    client.run("create . pkgb/1.0@")

    conanfile = textwrap.dedent("""
        [requires]
        pkgb/1.0

        [generators]
        pkg_config
        """)
    client.save({"conanfile.txt": conanfile}, clean_first=True)
    client.run("install .")
    pc_files = [os.path.basename(i) for i in glob.glob(os.path.join(client.current_folder, '*.pc'))]
    pc_files.sort()
    # Let's check all the PC file names created just in case
    assert pc_files == ['OpenCL.pc', 'OtherCL.pc', 'pkgb.pc']
    pc_content = client.load("OpenCL.pc")
    assert "Name: OpenCL-OpenCL" in pc_content
    assert "Description: Conan package: OpenCL-OpenCL" in pc_content
    assert "Requires:" not in pc_content
    pc_content = client.load("pkgb.pc")
    assert "Requires: OpenCL" in get_requires_from_content(pc_content)


@pytest.mark.skipif(platform.system() == "Windows", reason="Requires pkg-config")
@pytest.mark.tool_pkg_config
def test_pkg_config_definitions_escape():
    client = TestClient(path_with_spaces=False)
    conanfile = textwrap.dedent(r'''
        from conans import ConanFile
        class HelloLib(ConanFile):
            def package_info(self):
                self.cpp_info.defines.append("USER_CONFIG=\"user_config.h\"")
                self.cpp_info.defines.append('OTHER="other.h"')
                self.cpp_info.cflags.append("flag1=\"my flag1\"")
                self.cpp_info.cxxflags.append('flag2="my flag2"')
        ''')
    client.save({"conanfile.py": conanfile})
    client.run("export . hello/1.0@")
    client.save({"conanfile.txt": "[requires]\nhello/1.0\n"}, clean_first=True)
    client.run("install . --build=missing -g pkg_config")
    client.run_command("PKG_CONFIG_PATH=$(pwd) pkg-config --cflags hello")
    assert r'flag2=\"my flag2\" flag1=\"my flag1\" -DUSER_CONFIG=\"user_config.h\" -DOTHER=\"other.h\"' in client.out



import pytest
from conans.test.assets.genconanfile import GenConanfile
from conans.test.utils.tools import TestClient


@pytest.fixture(scope="module")
def build_all():
    """ Build a simple graph to test --build option

        foobar <- bar <- foo
               <--------|

        All packages are built from sources to keep a cache.
    :return: TestClient instance
    """
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_setting("build_type")})
    client.run("export . foo/1.0@user/testing")
    client.save({"conanfile.py": GenConanfile().with_require("foo/1.0@user/testing")
                .with_setting("build_type")})
    client.run("export . bar/1.0@user/testing")
    client.save({"conanfile.py": GenConanfile().with_require("foo/1.0@user/testing")
                .with_require("bar/1.0@user/testing")
                .with_setting("build_type")})
    client.run("export . foobar/1.0@user/testing")
    client.run("install foobar/1.0@user/testing --build")

    return client


def check_if_build_from_sources(refs_modes, output):
    for ref, mode in refs_modes.items():
        if mode == "Build":
            assert "{}/1.0@user/testing: Forced build from source".format(ref) in output
        else:
            assert "{}/1.0@user/testing: Forced build from source".format(ref) not in output


def test_install_build_single(build_all):
    """ When only --build=<ref> is passed, only <ref> must be built
    """
    build_all.run("install foobar/1.0@user/testing --build=foo")

    assert "bar/1.0@user/testing:7839863d5a059fc6579f28026763e1021268c55e - Cache" in build_all.out
    assert "foo/1.0@user/testing:4024617540c4f240a6a5e8911b0de9ef38a11a72 - Build" in build_all.out
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - Cache" in build_all.out
    assert "foo/1.0@user/testing: Forced build from source" in build_all.out
    assert "bar/1.0@user/testing: Forced build from source" not in build_all.out
    assert "foobar/1.0@user/testing: Forced build from source" not in build_all.out
    assert "No package matching" not in build_all.out


def test_install_build_double(build_all):
    """ When both --build=<ref1> and --build=<ref2> are passed, only both should be built
    """
    build_all.run("install foobar/1.0@user/testing --build=foo --build=bar")

    assert "bar/1.0@user/testing:7839863d5a059fc6579f28026763e1021268c55e - Build" in build_all.out
    assert "foo/1.0@user/testing:4024617540c4f240a6a5e8911b0de9ef38a11a72 - Build" in build_all.out
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - Cache" in build_all.out
    assert "foo/1.0@user/testing: Forced build from source" in build_all.out
    assert "bar/1.0@user/testing: Forced build from source" in build_all.out
    assert "foobar/1.0@user/testing: Forced build from source" not in build_all.out
    assert "No package matching" not in build_all.out


@pytest.mark.parametrize("build_arg,mode", [("--build", "Build"),
                                            ("--build=", "Cache"),
                                            ("--build=*", "Build")])
def test_install_build_only(build_arg, mode, build_all):
    """ When only --build is passed, all packages must be built from sources
        When only --build= is passed, it's considered an error
        When only --build=* is passed, all packages must be built from sources
    """
    build_all.run("install foobar/1.0@user/testing {}".format(build_arg))

    assert "bar/1.0@user/testing:7839863d5a059fc6579f28026763e1021268c55e - {}".format(mode) in build_all.out
    assert "foo/1.0@user/testing:4024617540c4f240a6a5e8911b0de9ef38a11a72 - {}".format(mode) in build_all.out
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - {}".format(mode) in build_all.out

    if "Build" == mode:
        assert "foo/1.0@user/testing: Forced build from source" in build_all.out
        assert "bar/1.0@user/testing: Forced build from source" in build_all.out
        assert "foobar/1.0@user/testing: Forced build from source" in build_all.out
        assert "No package matching" not in build_all.out
    else:
        assert "foo/1.0@user/testing: Forced build from source" not in build_all.out
        assert "bar/1.0@user/testing: Forced build from source" not in build_all.out
        assert "foobar/1.0@user/testing: Forced build from source" not in build_all.out
        assert "No package matching" in build_all.out


@pytest.mark.parametrize("build_arg,bar,foo,foobar", [("--build", "Cache", "Build", "Cache"),
                                                      ("--build=", "Cache", "Build", "Cache"),
                                                      ("--build=*", "Build", "Build", "Build")])
def test_install_build_all_with_single(build_arg, bar, foo, foobar, build_all):
    """ When --build is passed with another package, only the package must be built from sources.
        When --build= is passed with another package, only the package must be built from sources.
        When --build=* is passed with another package, all packages must be built from sources.
    """
    build_all.run("install foobar/1.0@user/testing --build=foo {}".format(build_arg))

    assert "bar/1.0@user/testing:7839863d5a059fc6579f28026763e1021268c55e - {}".format(bar) in build_all.out
    assert "foo/1.0@user/testing:4024617540c4f240a6a5e8911b0de9ef38a11a72 - {}".format(foo) in build_all.out
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - {}".format(foobar) in build_all.out
    check_if_build_from_sources({"foo": foo, "bar": bar, "foobar": foobar}, build_all.out)


@pytest.mark.parametrize("build_arg,bar,foo,foobar", [("--build", "Cache", "Cache", "Cache"),
                                                      ("--build=", "Cache", "Cache", "Cache"),
                                                      ("--build=*", "Build", "Cache", "Build")])
def test_install_build_all_with_single_skip(build_arg, bar, foo, foobar, build_all):
    """ When --build is passed with a skipped package, not all packages must be built from sources.
        When --build= is passed with another package, only the package must be built from sources.
        When --build=* is passed with another package, not all packages must be built from sources.

        The arguments order matter, that's why we need to run twice.
    """
    for argument in ["--build=!foo {}".format(build_arg),
                     "{} --build=!foo".format(build_arg)]:
        build_all.run("install foobar/1.0@user/testing {}".format(argument))
        assert "bar/1.0@user/testing:7839863d5a059fc6579f28026763e1021268c55e - {}".format(bar) in build_all.out
        assert "foo/1.0@user/testing:4024617540c4f240a6a5e8911b0de9ef38a11a72 - {}".format(foo) in build_all.out
        assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - {}".format(foobar) in build_all.out
        check_if_build_from_sources({"foo": foo, "bar": bar, "foobar": foobar}, build_all.out)


@pytest.mark.parametrize("build_arg,bar,foo,foobar", [("--build", "Cache", "Cache", "Cache"),
                                                      ("--build=", "Cache", "Cache", "Cache"),
                                                      ("--build=*", "Cache", "Cache", "Build")])
def test_install_build_all_with_double_skip(build_arg, bar, foo, foobar, build_all):
    """ When --build is passed with a skipped package, not all packages must be built from sources.
        When --build= is passed with another package, only the package must be built from sources.
        When --build=* is passed with another package, not all packages must be built from sources.

        The arguments order matter, that's why we need to run twice.
    """
    for argument in ["--build=!foo --build=!bar {}".format(build_arg),
                     "{} --build=!foo --build=!bar".format(build_arg)]:
        build_all.run("install foobar/1.0@user/testing {}".format(argument))

        assert "bar/1.0@user/testing:7839863d5a059fc6579f28026763e1021268c55e - {}".format(bar) in build_all.out
        assert "foo/1.0@user/testing:4024617540c4f240a6a5e8911b0de9ef38a11a72 - {}".format(foo) in build_all.out
        assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - {}".format(foobar) in build_all.out


def test_report_matches(build_all):
    """ When a wrong reference is passed to be build, an error message should be shown
    """
    build_all.run("install foobar/1.0@user/testing --build=* --build=baz")
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - Build" in build_all.out
    assert "No package matching 'baz' pattern found." in build_all.out

    build_all.run("install foobar/1.0@user/testing --build=* --build=!baz")
    assert "No package matching 'baz' pattern found." in build_all.out
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - Build" in build_all.out

    build_all.run("install foobar/1.0@user/testing --build=* --build=!baz --build=blah")
    assert "No package matching 'blah' pattern found." in build_all.out
    assert "No package matching 'baz' pattern found." in build_all.out
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - Build" in build_all.out

    build_all.run("install foobar/1.0@user/testing --build=* --build=!baz --build=!blah")
    assert "No package matching 'blah' pattern found." in build_all.out
    assert "No package matching 'baz' pattern found." in build_all.out
    assert "foobar/1.0@user/testing:89636fbae346e3983af2dd63f2c5246505e74be7 - Build" in build_all.out



import os
import re
import textwrap

import pytest
import six

from conans.test.utils.scm import create_local_git_repo, git_add_changes_commit, git_create_bare_repo
from conans.test.utils.test_files import temp_folder
from conans.test.utils.tools import TestClient
from conans.util.files import rmdir, save_files


@pytest.mark.skipif(six.PY2, reason="Only Py3")
class TestGitBasicCapture:
    """ base Git capture operations. They do not raise (unless errors)
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Git

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, self.recipe_folder)
                commit = git.get_commit()
                url = git.get_remote_url()
                self.output.info("URL: {}".format(url))
                self.output.info("COMMIT: {}".format(commit))
                in_remote = git.commit_in_remote(commit)
                self.output.info("COMMIT IN REMOTE: {}".format(in_remote))
                self.output.info("DIRTY: {}".format(git.is_dirty()))
        """)

    def test_capture_commit_local(self):
        """
        A local repo, without remote, will have commit, but no URL
        """
        c = TestClient()
        c.save({"conanfile.py": self.conanfile})
        commit = c.init_git_repo()
        c.run("export .")
        assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: URL: None" in c.out
        assert "pkg/0.1: COMMIT IN REMOTE: False" in c.out
        assert "pkg/0.1: DIRTY: False" in c.out

    def test_capture_remote_url(self):
        """
        a cloned repo, will have a default "origin" remote and will manage to get URL
        """
        folder = temp_folder()
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(folder))
        with c.chdir("myclone"):
            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: True" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out

    def test_capture_remote_pushed_commit(self):
        """
        a cloned repo, after doing some new commit, no longer commit in remote, until push
        """
        url = git_create_bare_repo()

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(url))
        with c.chdir("myclone"):
            c.save({"conanfile.py": self.conanfile + "\n# some coment!"})
            new_commit = git_add_changes_commit(c.current_folder)

            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(new_commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: False" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out
            c.run_command("git push")
            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(new_commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: True" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out


@pytest.mark.skipif(six.PY2, reason="Only Py3")
class TestGitCaptureSCM:
    """ test the get_url_and_commit() high level method intended for SCM capturing
    into conandata.yaml
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Git

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, self.recipe_folder)
                scm_url, scm_commit = git.get_url_and_commit()
                self.output.info("SCM URL: {}".format(scm_url))
                self.output.info("SCM COMMIT: {}".format(scm_commit))
        """)

    def test_capture_commit_local(self):
        """
        A local repo, without remote, will provide its own URL to the export(),
        and if it has local changes, it will be marked as dirty, and raise an error
        """
        c = TestClient()
        c.save({"conanfile.py": self.conanfile})
        commit = c.init_git_repo()
        c.run("export .")
        assert "This revision will not be buildable in other computer" in c.out
        assert "pkg/0.1: SCM COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: SCM URL: {}".format(c.current_folder.replace("\\", "/")) in c.out

        c.save({"conanfile.py": self.conanfile + "\n# something...."})
        c.run("export .", assert_error=True)
        assert "Repo is dirty, cannot capture url and commit" in c.out

    def test_capture_remote_url(self):
        """
        a cloned repo that is expored, will report the URL of the remote
        """
        folder = temp_folder()
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(folder))
        with c.chdir("myclone"):
            c.run("export .")
            assert "pkg/0.1: SCM COMMIT: {}".format(commit) in c.out
            assert "pkg/0.1: SCM URL: {}".format(url) in c.out

    def test_capture_remote_pushed_commit(self):
        """
        a cloned repo, after doing some new commit, no longer commit in remote, until push
        """
        url = git_create_bare_repo()

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(url))
        with c.chdir("myclone"):
            c.save({"conanfile.py": self.conanfile + "\n# some coment!"})
            new_commit = git_add_changes_commit(c.current_folder)

            c.run("export .")
            assert "This revision will not be buildable in other computer" in c.out
            assert "pkg/0.1: SCM COMMIT: {}".format(new_commit) in c.out
            # NOTE: commit not pushed yet, so locally is the current folder
            assert "pkg/0.1: SCM URL: {}".format(c.current_folder.replace("\\", "/")) in c.out
            c.run_command("git push")
            c.run("export .")
            assert "pkg/0.1: SCM COMMIT: {}".format(new_commit) in c.out
            assert "pkg/0.1: SCM URL: {}".format(url) in c.out


@pytest.mark.skipif(six.PY2, reason="Only Py3")
class TestGitBasicClone:
    """ base Git cloning operations
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def layout(self):
                self.folders.source = "source"

            def source(self):
                git = Git(self)
                git.clone(url="{url}", target=".")
                git.checkout(commit="{commit}")
                self.output.info("MYCMAKE: {{}}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {{}}".format(load(self, "src/myfile.h")))
        """)

    def test_clone_checkout(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)
        # This second commit will NOT be used, as I will use the above commit in the conanfile
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        c = TestClient()
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit)})
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

        # It also works in local flow
        c.run("source .")
        assert "conanfile.py (pkg/0.1): MYCMAKE: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE: myheader!" in c.out
        assert c.load("source/src/myfile.h") == "myheader!"
        assert c.load("source/CMakeLists.txt") == "mycmake"


class TestGitCloneWithArgs:
    """ Git cloning passing additional arguments
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def layout(self):
                self.folders.source = "source"

            def source(self):
                git = Git(self)
                git.clone(url="{url}", target=".", args={args})
                self.output.info("MYCMAKE: {{}}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {{}}".format(load(self, "src/myfile.h")))
        """)

    def test_clone_specify_branch_or_tag(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder,
                                                   commits=3, branch="main", tags=["v1.2.3"])

        c = TestClient()
        git_args = ['--branch', 'main']
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit, args=str(git_args))})
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

        git_args = ['--branch', 'v1.2.3']
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit, args=str(git_args))})
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

    def test_clone_invalid_branch_argument(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder,
                                                   commits=3, branch="main", tags=["v1.2.3"])
        c = TestClient()
        git_args = ['--branch', 'foobar']
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit, args=str(git_args))})
        with pytest.raises(Exception):
            c.run("create .")
            assert "Remote branch foobar not found" in c.out

@pytest.mark.skipif(six.PY2, reason="Only Py3")
class TestGitBasicSCMFlow:
    """ Build the full new SCM approach:
    - export() captures the URL and commit with get_url_and_commit(
    - export() stores it in conandata.yml
    - source() recovers the info from conandata.yml and clones it
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load, update_conandata

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, self.recipe_folder)
                scm_url, scm_commit = git.get_url_and_commit()
                update_conandata(self, {"sources": {"commit": scm_commit, "url": scm_url}})

            def layout(self):
                self.folders.source = "."

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                git.clone(url=sources["url"], target=".")
                git.checkout(commit=sources["commit"])
                self.output.info("MYCMAKE: {}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {}".format(load(self, "src/myfile.h")))

            def build(self):
                cmake = os.path.join(self.source_folder, "CMakeLists.txt")
                file_h = os.path.join(self.source_folder, "src/myfile.h")
                self.output.info("MYCMAKE-BUILD: {}".format(load(self, cmake)))
                self.output.info("MYFILE-BUILD: {}".format(load(self, file_h)))
        """)

    def test_full_scm(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile,
                                                   "src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)

        c = TestClient(default_server_user=True)
        c.run_command('git clone "{}" .'.format(url))
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out
        c.run("upload * --all -c")

        # Do a change and commit, this commit will not be used by package
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install pkg/0.1@ --build=pkg")
        assert "pkg/0.1: MYCMAKE: mycmake" in c2.out
        assert "pkg/0.1: MYFILE: myheader!" in c2.out

        # local flow
        c.run("install .")
        c.run("build .")
        assert "conanfile.py (pkg/0.1): MYCMAKE-BUILD: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE-BUILD: myheader!" in c.out

    def test_branch_flow(self):
        """ Testing that when a user creates a branch, and pushes a commit,
        the package can still be built from sources, and get_url_and_commit() captures the
        remote URL and not the local
        """
        url = git_create_bare_repo()
        c = TestClient(default_server_user=True)
        c.run_command('git clone "{}" .'.format(url))
        c.save({"conanfile.py": self.conanfile,
                "src/myfile.h": "myheader!",
                "CMakeLists.txt": "mycmake"})
        c.run_command("git checkout -b mybranch")
        git_add_changes_commit(folder=c.current_folder)
        c.run_command("git push --set-upstream origin mybranch")
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out
        c.run("upload * --all -c")
        rmdir(c.current_folder)  # Remove current folder to make sure things are not used from here

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install pkg/0.1@ --build=pkg")
        assert "pkg/0.1: MYCMAKE: mycmake" in c2.out
        assert "pkg/0.1: MYFILE: myheader!" in c2.out


@pytest.mark.skipif(six.PY2, reason="Only Py3")
class TestGitBasicSCMFlowSubfolder:
    """ Same as above, but conanfile.py put in "conan" subfolder in the root
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load, update_conandata

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, os.path.dirname(self.recipe_folder)) # PARENT!
                scm_url, scm_commit = git.get_url_and_commit()
                update_conandata(self, {"sources": {"commit": scm_commit, "url": scm_url}})

            def layout(self):
                self.folders.root = ".."
                self.folders.source = "."

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                git.clone(url=sources["url"], target=".")
                git.checkout(commit=sources["commit"])
                self.output.info("MYCMAKE: {}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {}".format(load(self, "src/myfile.h")))

            def build(self):
                cmake = os.path.join(self.source_folder, "CMakeLists.txt")
                file_h = os.path.join(self.source_folder, "src/myfile.h")
                self.output.info("MYCMAKE-BUILD: {}".format(load(self, cmake)))
                self.output.info("MYFILE-BUILD: {}".format(load(self, file_h)))
        """)

    def test_full_scm(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"conan/conanfile.py": self.conanfile,
                                                   "src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)

        c = TestClient(default_server_user=True)
        c.run_command('git clone "{}" .'.format(url))
        c.run("create conan")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out
        c.run("upload * --all -c")

        # Do a change and commit, this commit will not be used by package
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install pkg/0.1@ --build=pkg")
        assert "pkg/0.1: MYCMAKE: mycmake" in c2.out
        assert "pkg/0.1: MYFILE: myheader!" in c2.out

        # local flow
        c.run("install conan")
        c.run("build conan")
        assert "conanfile.py (pkg/0.1): MYCMAKE-BUILD: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE-BUILD: myheader!" in c.out


@pytest.mark.skipif(six.PY2, reason="Only Py3")
class TestGitMonorepoSCMFlow:
    """ Build the full new SCM approach:
    Same as above but with a monorepo with multiple subprojects
    """
    # TODO: swap_child_folder() not documented, not public usage
    conanfile = textwrap.dedent("""
        import os, shutil
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load, update_conandata
        from conan.tools.files.files import swap_child_folder

        class Pkg(ConanFile):
            name = "{pkg}"
            version = "0.1"

            {requires}

            def export(self):
                git = Git(self, self.recipe_folder)
                scm_url, scm_commit = git.get_url_and_commit()
                self.output.info("CAPTURING COMMIT: {{}}!!!".format(scm_commit))
                folder = os.path.basename(self.recipe_folder)
                update_conandata(self, {{"sources": {{"commit": scm_commit, "url": scm_url,
                                                      "folder": folder}}}})

            def layout(self):
                self.folders.source = "."

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                git.clone(url=sources["url"], target=".")
                git.checkout(commit=sources["commit"])
                swap_child_folder(self.source_folder, sources["folder"])

            def build(self):
                cmake = os.path.join(self.source_folder, "CMakeLists.txt")
                file_h = os.path.join(self.source_folder, "src/myfile.h")
                self.output.info("MYCMAKE-BUILD: {{}}".format(load(self, cmake)))
                self.output.info("MYFILE-BUILD: {{}}".format(load(self, file_h)))
        """)

    def test_full_scm(self):
        folder = os.path.join(temp_folder(), "myrepo")
        conanfile1 = self.conanfile.format(pkg="pkg1", requires="")
        conanfile2 = self.conanfile.format(pkg="pkg2", requires="requires = 'pkg1/0.1'")
        url, commit = create_local_git_repo(files={"sub1/conanfile.py": conanfile1,
                                                   "sub1/src/myfile.h": "myheader1!",
                                                   "sub1/CMakeLists.txt": "mycmake1!",
                                                   "sub2/conanfile.py": conanfile2,
                                                   "sub2/src/myfile.h": "myheader2!",
                                                   "sub2/CMakeLists.txt": "mycmake2!"
                                                   },
                                            folder=folder)

        c = TestClient(default_server_user=True)
        c.run_command('git clone "{}" .'.format(url))
        c.run("create sub1")
        commit = re.search(r"CAPTURING COMMIT: (\S+)!!!", str(c.out)).group(1)
        assert "pkg1/0.1: MYCMAKE-BUILD: mycmake1!" in c.out
        assert "pkg1/0.1: MYFILE-BUILD: myheader1!" in c.out

        c.save({"sub2/src/myfile.h": "my2header!"})
        git_add_changes_commit(folder=c.current_folder)
        c.run("create sub2")
        assert "pkg2/0.1: MYCMAKE-BUILD: mycmake2!" in c.out
        assert "pkg2/0.1: MYFILE-BUILD: my2header!" in c.out

        # Exporting again sub1, gives us exactly the same revision as before
        c.run("export sub1")
        assert "CAPTURING COMMIT: {}".format(commit) in c.out
        c.run("upload * --all -c -r=default")

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install pkg2/0.1@ --build")
        assert "pkg1/0.1: Checkout: {}".format(commit) in c2.out
        assert "pkg1/0.1: MYCMAKE-BUILD: mycmake1!" in c2.out
        assert "pkg1/0.1: MYFILE-BUILD: myheader1!" in c2.out
        assert "pkg2/0.1: MYCMAKE-BUILD: mycmake2!" in c2.out
        assert "pkg2/0.1: MYFILE-BUILD: my2header!" in c2.out


class TestConanFileSubfolder:
    """verify that we can have a conanfile in a subfolder
        # https://github.com/conan-io/conan/issues/11275
    """

    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import update_conandata, load

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, os.path.dirname(self.recipe_folder))
                url, commit = git.get_url_and_commit()
                # We store the current url and commit in conandata.yml
                update_conandata(self, {"sources": {"commit": commit, "url": url}})
                self.output.info("URL: {}".format(url))
                self.output.info("COMMIT: {}".format(commit))

            def layout(self):
                pass # self.folders.source = "source"

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                url = sources["url"]
                commit = sources["commit"]
                git.clone(url=url, target=".")
                git.checkout(commit=commit)
                self.output.info("MYCMAKE: {}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {}".format(load(self, "src/myfile.h")))
        """)

    def test_conanfile_subfolder(self):
        """
        A local repo, without remote, will have commit, but no URL
        """
        c = TestClient()
        c.save({"conan/conanfile.py": self.conanfile,
                "CMakeLists.txt": "mycmakelists",
                "src/myfile.h": "myheader"})
        commit = c.init_git_repo()
        c.run("export conan")
        assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: URL: {}".format(c.current_folder.replace("\\", "/")) in c.out

        c.run("create conan")
        assert "pkg/0.1: MYCMAKE: mycmakelists" in c.out
        assert "pkg/0.1: MYFILE: myheader" in c.out



import os
import platform
import re
import textwrap

import pytest

from conans.model.ref import ConanFileReference, PackageReference
from conans.test.assets.genconanfile import GenConanfile
from conans.test.assets.pkg_cmake import pkg_cmake
from conans.test.utils.tools import TestClient


def test_local_static_generators_folder():
    """If we configure a generators folder in the layout, the generator files:
      - If belong to new generators: go to the specified folder: "my_generators"
      - If belong to old generators or txt: remains in the install folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type"))
    conan_file += """
    generators = "cmake", "CMakeToolchain"
    def layout(self):
        self.folders.build = "build-{}".format(self.settings.build_type)
        self.folders.generators = "{}/generators".format(self.folders.build)
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install")

    old_install_folder = os.path.join(client.current_folder, "my_install")
    conaninfo = os.path.join(old_install_folder, "conaninfo.txt")
    conanbuildinfo = os.path.join(old_install_folder, "conanbuildinfo.txt")
    cmake_generator_path = os.path.join(old_install_folder, "conanbuildinfo.cmake")
    cmake_toolchain_generator_path = os.path.join(old_install_folder, "conan_toolchain.cmake")
    assert os.path.exists(conaninfo)
    assert os.path.exists(conanbuildinfo)
    assert os.path.exists(cmake_generator_path)
    assert not os.path.exists(cmake_toolchain_generator_path)

    build_folder = os.path.join(client.current_folder, "build-Release")
    generators_folder = os.path.join(build_folder, "generators")
    conaninfo = os.path.join(generators_folder, "conaninfo.txt")
    conanbuildinfo = os.path.join(generators_folder, "conanbuildinfo.txt")
    cmake_generator_path = os.path.join(generators_folder, "conanbuildinfo.cmake")
    cmake_toolchain_generator_path = os.path.join(generators_folder, "conan_toolchain.cmake")
    assert not os.path.exists(conaninfo)
    assert not os.path.exists(conanbuildinfo)
    assert not os.path.exists(cmake_generator_path)
    assert os.path.exists(cmake_toolchain_generator_path)


def test_local_dynamic_generators_folder():
    """If we configure a generators folder in the layout, the generator files:
      - If belong to new generators: go to the specified folder: "my_generators"
      - "txt" and old ones always to the install folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type").
                     with_import("from conan.tools.cmake import CMakeToolchain, CMake"))
    conan_file += """
    def generate(self):
        tc = CMakeToolchain(self)
        tc.generate()

    def layout(self):
        self.folders.build = "build-{}".format(self.settings.build_type)
        self.folders.generators = "{}/generators".format(self.folders.build)
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install -g cmake")

    old_install_folder = os.path.join(client.current_folder, "my_install")
    conaninfo = os.path.join(old_install_folder, "conaninfo.txt")
    conanbuildinfo = os.path.join(old_install_folder, "conanbuildinfo.txt")
    cmake_generator_path = os.path.join(old_install_folder, "conanbuildinfo.cmake")
    cmake_toolchain_generator_path = os.path.join(old_install_folder, "conan_toolchain.cmake")
    assert os.path.exists(conaninfo)
    assert os.path.exists(conanbuildinfo)
    assert os.path.exists(cmake_generator_path)
    assert not os.path.exists(cmake_toolchain_generator_path)

    build_folder = os.path.join(client.current_folder, "build-Release")
    generators_folder = os.path.join(build_folder, "generators")
    conaninfo = os.path.join(generators_folder, "conaninfo.txt")
    conanbuildinfo = os.path.join(generators_folder, "conanbuildinfo.txt")
    cmake_generator_path = os.path.join(generators_folder, "conanbuildinfo.cmake")
    cmake_toolchain_generator_path = os.path.join(generators_folder, "conan_toolchain.cmake")
    assert not os.path.exists(conaninfo)
    assert not os.path.exists(conanbuildinfo)
    assert not os.path.exists(cmake_generator_path)
    assert os.path.exists(cmake_toolchain_generator_path)


def test_no_layout_generators_folder():
    """If we don't configure a generators folder in the layout, the generator files:
      - all go to the install_folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type").
                     with_import("from conan.tools.cmake import CMakeToolchain, CMake"))
    conan_file += """
    def generate(self):
        tc = CMakeToolchain(self)
        tc.generate()
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install -g cmake")

    old_install_folder = os.path.join(client.current_folder, "my_install")
    conaninfo = os.path.join(old_install_folder, "conaninfo.txt")
    conanbuildinfo = os.path.join(old_install_folder, "conanbuildinfo.txt")
    cmake_generator_path = os.path.join(old_install_folder, "conanbuildinfo.cmake")
    cmake_toolchain_generator_path = os.path.join(old_install_folder, "conan_toolchain.cmake")
    assert os.path.exists(conaninfo)
    assert os.path.exists(conanbuildinfo)
    assert os.path.exists(cmake_generator_path)
    # In the install_folder
    assert os.path.exists(cmake_toolchain_generator_path)

    # But not in the base folder
    assert not os.path.exists(os.path.join(client.current_folder, "conan_toolchain.cmake"))


def test_local_build():
    """If we configure a build folder in the layout, the installed files in a "conan build ."
    go to the specified folder: "my_build"
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """

    def layout(self):
        self.folders.generators = "my_generators"
        self.folders.build = "my_build"

    def build(self):
        self.output.warn("Generators folder: {}".format(self.folders.generators_folder))
        tools.save("build_file.dll", "bar")

"""
    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install")
    # FIXME: This should change to "build ." when "conan build" computes the graph
    client.run("build . -if=my_install")
    dll = os.path.join(client.current_folder, "my_build", "build_file.dll")
    assert os.path.exists(dll)


def test_local_build_change_base():
    """If we configure a build folder in the layout, the build files in a "conan build ."
    go to the specified folder: "my_build under the modified base one "common"
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """
    def layout(self):
        self.folders.build = "my_build"
    def build(self):
        tools.save("build_file.dll", "bar")
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -if=common -bf=common")
    client.run("build . -if=common -bf=common")
    dll = os.path.join(client.current_folder, "common", "my_build", "build_file.dll")
    assert os.path.exists(dll)


def test_local_source():
    """If we configure a source folder in the layout, the downloaded files in a "conan source ."
    DON'T go to the specified folder: "my_source" but to the root source folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """
    def layout(self):
        self.folders.source = "my_source"

    def source(self):
        tools.save("downloaded.h", "bar")
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install")
    # FIXME: This should change to "source ." when "conan source" computes the graph
    client.run("source . -if=my_install")
    header = os.path.join(client.current_folder, "my_source", "downloaded.h")
    assert os.path.exists(header)


def test_local_source_change_base():
    """If we configure a source folder in the layout, the source files in a "conan source ."
    DON'T go to the specified folder: "my_source under the modified base one "all_source"
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """
    def layout(self):
        self.folders.source = "my_source"

    def source(self):
        tools.save("downloaded.h", "bar")
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -if=common")
    client.run("source . -if=common -sf=common")
    header = os.path.join(client.current_folder, "common", "my_source", "downloaded.h")
    assert os.path.exists(header)


def test_export_pkg():
    """The export-pkg, calling the "package" method, follows the layout if `cache_package_layout` """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """
    no_copy_source = True

    def layout(self):
        self.folders.source = "my_source"
        self.folders.build = "my_build"

    def source(self):
        tools.save("downloaded.h", "bar")

    def build(self):
        tools.save("library.lib", "bar")
        tools.save("generated.h", "bar")

    def package(self):
        self.output.warn("Source folder: {}".format(self.source_folder))
        self.output.warn("Build folder: {}".format(self.build_folder))
        self.output.warn("Package folder: {}".format(self.package_folder))
        self.copy("*.h")
        self.copy("*.lib")
    """

    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install")
    client.run("source . -if=my_install")
    client.run("build . -if=my_install")
    client.run("export-pkg . lib/1.0@ -if=my_install")
    package_id = re.search(r"lib/1.0: Package '(\S+)' created", str(client.out)).group(1)
    ref = ConanFileReference.loads("lib/1.0@")
    pref = PackageReference(ref, package_id)
    sf = os.path.join(client.current_folder, "my_source")
    bf = os.path.join(client.current_folder, "my_build")
    pf = client.cache.package_layout(ref).package(pref)

    assert "WARN: Source folder: {}".format(sf) in client.out
    assert "WARN: Build folder: {}".format(bf) in client.out
    assert "WARN: Package folder: {}".format(pf) in client.out

    # Check the artifacts packaged
    assert os.path.exists(os.path.join(pf, "generated.h"))
    assert os.path.exists(os.path.join(pf, "library.lib"))


@pytest.mark.xfail(reason="We cannot test the export-pkg from a local package folder when we cannot "
                          "run the package method")
def test_export_pkg_local():
    """The export-pkg, without calling "package" method, with local package, follows the layout"""
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """
    no_copy_source = True

    def layout(self):
        self.folders.source = "my_source"
        self.folders.build = "my_build"
        self.folders.package = "my_package"

    def source(self):
        tools.save("downloaded.h", "bar")

    def build(self):
        tools.save("library.lib", "bar")
        tools.save("generated.h", "bar")

    def package(self):
        self.output.warn("Source folder: {}".format(self.source_folder))
        self.output.warn("Build folder: {}".format(self.build_folder))
        self.output.warn("Package folder: {}".format(self.package_folder))
        self.copy("*.h")
        self.copy("*.lib")
    """

    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install")
    client.run("source . -if=my_install")
    client.run("build . -if=my_install")
    client.run("package . -if=my_install")
    sf = os.path.join(client.current_folder, "my_source")
    bf = os.path.join(client.current_folder, "my_build")
    pf = os.path.join(client.current_folder, "my_package")
    assert "WARN: Source folder: {}".format(sf) in client.out
    assert "WARN: Build folder: {}".format(bf) in client.out
    assert "WARN: Package folder: {}".format(pf) in client.out

    client.run("export-pkg . lib/1.0@ -if=my_install -pf=my_package")
    ref = ConanFileReference.loads("lib/1.0@")
    pref = PackageReference(ref, "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9")
    pf_cache = client.cache.package_layout(ref).package(pref)

    # Check the artifacts packaged, THERE IS NO "my_package" in the cache
    assert "my_package" not in pf_cache
    assert os.path.exists(os.path.join(pf_cache, "generated.h"))
    assert os.path.exists(os.path.join(pf_cache, "library.lib"))

    # Doing a conan create: Same as export-pkg, there is "my_package" in the cache
    client.run("create . lib/1.0@")
    ref = ConanFileReference.loads("lib/1.0@")
    pref = PackageReference(ref, "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9")
    pf_cache = client.cache.package_layout(ref).package(pref)
    assert "my_package" not in pf_cache
    assert os.path.exists(os.path.join(pf_cache, "generated.h"))
    assert os.path.exists(os.path.join(pf_cache, "library.lib"))


def test_imports():
    """The 'conan imports' follows the layout"""
    client = TestClient()
    # Hello to be reused
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """
    no_copy_source = True

    def build(self):
        tools.save("library.dll", "bar")
        tools.save("generated.h", "bar")

    def package(self):
        self.copy("*.h")
        self.copy("*.dll")
    """
    client.save({"conanfile.py": conan_file})
    client.run("create . hello/1.0@")

    # Consumer of the hello importing the shared
    conan_file = str(GenConanfile().with_import("from conans import tools"))
    conan_file += """
    no_copy_source = True
    requires = "hello/1.0"

    def layout(self):
        self.folders.imports = "my_imports"

    def imports(self):
        self.output.warn("Imports folder: {}".format(self.imports_folder))
        self.copy("*.dll")

    """

    client.save({"conanfile.py": conan_file})
    client.run("install . -if=my_install")
    client.run("imports . -if=my_install")

    imports_folder = os.path.join(client.current_folder, "my_imports")
    dll_path = os.path.join(imports_folder, "library.dll")

    assert "WARN: Imports folder: {}".format(imports_folder) in client.out
    assert os.path.exists(dll_path)

    # If we do a conan create, the imports folder is used also in the cache
    client.run("create . foo/1.0@")
    package_id = re.search(r"foo/1.0:(\S+)", str(client.out)).group(1)
    ref = ConanFileReference.loads("foo/1.0@")
    pref = PackageReference(ref, package_id)
    bfolder = client.cache.package_layout(ref).build(pref)
    imports_folder = os.path.join(bfolder, "my_imports")
    assert "WARN: Imports folder: {}".format(imports_folder) in client.out


def test_start_dir_failure():
    c = TestClient()
    c.save(pkg_cmake("dep", "0.1"))
    c.run("install .")
    expected_path = os.path.join(c.current_folder, "build", "generators", "conan_toolchain.cmake")
    assert os.path.exists(expected_path)
    os.unlink(expected_path)
    with c.chdir("build"):
        c.run("install ..")
    assert os.path.exists(expected_path)


def test_importdir_failure():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conans import ConanFile
        from conan.tools.files import save
        class Cosumer(ConanFile):
            requires = "dep/0.1"
            settings = "build_type"
            def layout(self):
                self.folders.imports = "Import" + str(self.settings.build_type)
                self.folders.build = "build"
            def build(self):
                self.output.info("saving file")
                save(self, "mybuild.txt", "mybuild")
            def imports(self):
                self.copy("myfile.txt")
        """)
    c.save({"dep/conanfile.py": GenConanfile().with_package_file("myfile.txt", "mycontent"),
            "consumer/conanfile.py": conanfile})
    c.run("create dep dep/0.1@")
    with c.chdir("consumer"):
        c.run("install . -s build_type=Release")
        expected_path_release = os.path.join(c.current_folder, "ImportRelease", "myfile.txt")
        assert os.path.exists(expected_path_release)
        c.run("build .")
        expected_build_file = os.path.join(c.current_folder, "build", "mybuild.txt")
        assert os.path.exists(expected_build_file)
        c.run("install . -s build_type=Debug")
        expected_path_debug = os.path.join(c.current_folder, "ImportDebug", "myfile.txt")
        assert os.path.exists(expected_path_debug)

        os.unlink(expected_path_release)
        os.unlink(expected_path_debug)
        os.unlink(expected_build_file)
        with c.chdir("build"):
            c.run("install .. -s build_type=Release")
            assert os.path.exists(expected_path_release)
            c.run("build ..")
            assert os.path.exists(expected_build_file)
            c.run("install .. -s build_type=Debug")
            assert os.path.exists(expected_path_debug)


def test_local_folders_without_layout():
    """ Test that the "conan install" can report the local source and build folder
    """
    # https://github.com/conan-io/conan/issues/10566
    client = TestClient()
    conan_file = textwrap.dedent("""
        from conan import ConanFile
        class Test(ConanFile):
            version = "1.2.3"
            name = "test"

            def layout(self):
                self.folders.source = "."
                self.folders.build = "build"

            def generate(self):
                self.output.info("generate sf: {}".format(self.source_folder))
                self.output.info("generate bf: {}".format(self.build_folder))
        """)

    client.save({"conanfile.py": conan_file})
    client.run("install .")
    assert "conanfile.py (test/1.2.3): generate sf: {}".format(client.current_folder) in client.out
    assert "conanfile.py (test/1.2.3): generate bf: {}".format(client.current_folder) in client.out



import os
import unittest

from conans.assets.templates import SEARCH_TABLE_HTML, INFO_GRAPH_DOT, INFO_GRAPH_HTML
from conans.client.tools import save
from conans.model.ref import ConanFileReference
from conans.test.utils.tools import TestClient, GenConanfile


class UserOverridesTemplatesTestCase(unittest.TestCase):
    lib_ref = ConanFileReference.loads("lib/version")
    app_ref = ConanFileReference.loads("app/version")

    @classmethod
    def setUpClass(cls):
        cls.t = TestClient()
        cls.t.save({'lib.py': GenConanfile().with_setting("os"),
                    'app.py': GenConanfile().with_setting("os").with_require(cls.lib_ref)})
        cls.t.run("create lib.py {}@ -s os=Windows".format(cls.lib_ref))
        cls.t.run("create lib.py {}@ -s os=Linux".format(cls.lib_ref))
        cls.t.run("create app.py {}@ -s os=Windows".format(cls.app_ref))
        cls.t.run("create app.py {}@ -s os=Linux".format(cls.app_ref))

    def test_table_html(self):
        table_template_path = os.path.join(self.t.cache_folder, 'templates', SEARCH_TABLE_HTML)
        save(table_template_path, content='{{ base_template_path }}')
        self.t.run("search {}@ --table=output.html".format(self.lib_ref))
        content = self.t.load("output.html")
        self.assertEqual(os.path.join(self.t.cache_folder, 'templates', 'output'), content)

    def test_graph_html(self):
        table_template_path = os.path.join(self.t.cache_folder, 'templates', INFO_GRAPH_HTML)
        save(table_template_path, content='{{ base_template_path }}')
        self.t.run("info {}@ --graph=output.html".format(self.app_ref))
        content = self.t.load("output.html")
        self.assertEqual(os.path.join(self.t.cache_folder, 'templates', 'output'), content)

    def test_graph_dot(self):
        table_template_path = os.path.join(self.t.cache_folder, 'templates', INFO_GRAPH_DOT)
        save(table_template_path, content='{{ base_template_path }}')
        self.t.run("info {}@ --graph=output.dot".format(self.app_ref))
        content = self.t.load("output.dot")
        self.assertEqual(os.path.join(self.t.cache_folder, 'templates', 'output'), content)



import textwrap

from conans.model.ref import ConanFileReference
from conans.test.utils.tools import TestClient


class TestExportFoldersAvailability:

    def test_export_sources_folder_availability_local_methods(self):
        conanfile = textwrap.dedent('''
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                self.output.info("Running generate, value {}!".format(self.export_sources_folder))

            def export(self):
                self.output.info("Running export, value {}!".format(self.export_sources_folder))

            def export_sources(self):
                self.output.info("Running export_sources, value {}!".format(self.export_sources_folder))

            def source(self):
                self.output.info("Running source, value {}!".format(self.export_sources_folder))

            def build(self):
                self.output.info("Running build, value {}!".format(self.export_sources_folder))

            def package(self):
                self.output.info("Running package, value {}!".format(self.export_sources_folder))

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})

        client.run("export . foo/1.0@")
        cache_exports_sources = client.cache.package_layout(ConanFileReference.loads("foo/1.0")).export_sources()
        assert "Running export, value None!" in client.out
        assert "Running export_sources, value {}!".format(cache_exports_sources) in client.out

        client.run("install .")
        # This might be a bit unexpected but self.source_folder is also defined in generate()
        # so it looks consistent with having declared the export_sources_folder
        assert "Running generate, value {}!".format(client.current_folder) in client.out

        client.run("source .")
        assert "Running source, value {}!".format(client.current_folder) in client.out

        client.run("build .")
        assert "Running build, value {}!".format(client.current_folder) in client.out

    def test_export_folder_availability_local_methods(self):
        conanfile = textwrap.dedent('''
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                self.output.info("Running generate, value {}!".format(self.export_folder))

            def export(self):
                self.output.info("Running export, value {}!".format(self.export_folder))

            def export_sources(self):
                self.output.info("Running export_sources, value {}!".format(self.export_folder))

            def source(self):
                self.output.info("Running source, value {}!".format(self.export_folder))

            def build(self):
                self.output.info("Running build, value {}!".format(self.export_folder))

            def package(self):
                self.output.info("Running package, value {}!".format(self.export_folder))

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})

        client.run("export . foo/1.0@")
        cache_exports = client.cache.package_layout(ConanFileReference.loads("foo/1.0")).export()
        assert "Running export_sources, value None!" in client.out
        assert "Running export, value {}!".format(cache_exports) in client.out

        client.run("install .")
        assert "Running generate, value None!" in client.out

        client.run("source .")
        assert "Running source, value None!" in client.out

        client.run("build .")
        assert "Running build, value None!" in client.out

    def test_export_folder_availability_create(self):
        conanfile = textwrap.dedent('''
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                self.output.info("Running generate, value {}!".format(self.export_folder))

            def export(self):
                self.output.info("Running export, value {}!".format(self.export_folder))

            def export_sources(self):
                self.output.info("Running export_sources, value {}!".format(self.export_folder))

            def source(self):
                self.output.info("Running source, value {}!".format(self.export_folder))

            def build(self):
                self.output.info("Running build, value {}!".format(self.export_folder))

            def package(self):
                self.output.info("Running package, value {}!".format(self.export_folder))

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        cache_exports = client.cache.package_layout(ConanFileReference.loads("foo/1.0")).export()
        client.run("create . foo/1.0@")
        assert "Running export_sources, value None!" in client.out
        assert "Running export, value {}!".format(cache_exports) in client.out
        assert "Running generate, value None!" in client.out
        assert "Running source, value None!" in client.out
        assert "Running build, value None!" in client.out

    def test_export_sources_folder_availability_create(self):
        conanfile = textwrap.dedent('''
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                self.output.info("Running generate, value {}!".format(self.export_sources_folder))

            def export(self):
                self.output.info("Running export, value {}!".format(self.export_sources_folder))

            def export_sources(self):
                self.output.info("Running export_sources, value {}!".format(self.export_sources_folder))

            def source(self):
                self.output.info("Running source, value {}!".format(self.export_sources_folder))

            def build(self):
                self.output.info("Running build, value {}!".format(self.export_sources_folder))

            def package(self):
                self.output.info("Running package, value {}!".format(self.export_sources_folder))

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        cache_exports = client.cache.package_layout(ConanFileReference.loads("foo/1.0")).export_sources()
        client.run("create . foo/1.0@")
        assert "Running export, value None!" in client.out
        assert "Running export_sources, value {}!".format(cache_exports) in client.out
        cache_base_source = client.cache.package_layout(ConanFileReference.loads("foo/1.0")).source()
        assert "Running generate, value {}!".format(cache_base_source) in client.out
        assert "Running source, value {}!".format(cache_base_source) in client.out
        assert "Running build, value {}!".format(cache_base_source) in client.out



import os
import platform
import subprocess
import textwrap

import pytest

from conan.tools.env import Environment
from conans.client.tools import chdir
from conans.test.utils.mocks import ConanFileMock
from conans.test.utils.test_files import temp_folder
from conans.util.files import save


@pytest.fixture
def env():
    env = Environment()
    env.define("MyVar", "MyValue")
    env.define("MyVar1", "MyValue1")
    env.append("MyVar2", "MyValue2")
    env.prepend("MyVar3", "MyValue3")
    env.unset("MyVar4")
    env.define("MyVar5", "MyValue5 With Space5=More Space5;:More")
    env.append("MyVar6", "MyValue6")  # Append, but previous not existing
    env.define_path("MyPath1", "/Some/Path1/")
    env.append_path("MyPath2", ["/Some/Path2/", "/Other/Path2/"])
    env.prepend_path("MyPath3", "/Some/Path3/")
    env.unset("MyPath4")
    return env


@pytest.fixture
def prevenv():
    return {
        "MyVar1": "OldVar1",
        "MyVar2": "OldVar2",
        "MyVar3": "OldVar3 with spaces",
        "MyVar4": "OldVar4 with (some) special @ characters",
        "MyPath1": "OldPath1",
        "MyPath2": "OldPath2",
        "MyPath3": "OldPath3",
        "MyPath4": "OldPath4",
    }


def check_env_files_output(cmd_, prevenv):
    result, _ = subprocess.Popen(cmd_, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                 env=prevenv, shell=True).communicate()
    out = result.decode()

    assert "MyVar=MyValue!!" in out
    assert "MyVar1=MyValue1!!" in out
    assert "MyVar2=OldVar2 MyValue2!!" in out
    assert "MyVar3=MyValue3 OldVar3 with spaces!!" in out
    assert "MyVar4=!!" in out
    assert "MyVar5=MyValue5 With Space5=More Space5;:More!!" in out
    assert "MyVar6= MyValue6!!" in out  # The previous is non existing, append has space
    assert "MyPath1=/Some/Path1/!!" in out
    assert os.pathsep.join(["MyPath2=OldPath2", "/Some/Path2/", "/Other/Path2/!!"]) in out
    assert os.pathsep.join(["MyPath3=/Some/Path3/", "OldPath3!!"]) in out
    assert "MyPath4=!!" in out

    # This should be output when deactivated
    assert "MyVar=!!" in out
    assert "MyVar1=OldVar1!!" in out
    assert "MyVar2=OldVar2!!" in out
    assert "MyVar3=OldVar3 with spaces!!" in out
    assert "MyVar4=OldVar4 with (some) special @ characters!!" in out
    assert "MyVar5=!!" in out
    assert "MyVar6=!!" in out
    assert "MyPath1=OldPath1!!" in out
    assert "MyPath2=OldPath2!!" in out
    assert "MyPath3=OldPath3!!" in out
    assert "MyPath4=OldPath4!!" in out


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_env_files_bat(env, prevenv):
    display = textwrap.dedent("""\
        @echo off
        echo MyVar=%MyVar%!!
        echo MyVar1=%MyVar1%!!
        echo MyVar2=%MyVar2%!!
        echo MyVar3=%MyVar3%!!
        echo MyVar4=%MyVar4%!!
        echo MyVar5=%MyVar5%!!
        echo MyVar6=%MyVar6%!!
        echo MyPath1=%MyPath1%!!
        echo MyPath2=%MyPath2%!!
        echo MyPath3=%MyPath3%!!
        echo MyPath4=%MyPath4%!!
        """)
    t = temp_folder()
    with chdir(t):
        env = env.vars(ConanFileMock())
        env.save_bat("test.bat")
        save("display.bat", display)
        cmd = "test.bat && display.bat && deactivate_test.bat && display.bat"
        check_env_files_output(cmd, prevenv)


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_env_files_ps1(env, prevenv):
    prevenv.update(dict(os.environ.copy()))

    display = textwrap.dedent("""\
        echo "MyVar=$env:MyVar!!"
        echo "MyVar1=$env:MyVar1!!"
        echo "MyVar2=$env:MyVar2!!"
        echo "MyVar3=$env:MyVar3!!"
        echo "MyVar4=$env:MyVar4!!"
        echo "MyVar5=$env:MyVar5!!"
        echo "MyVar6=$env:MyVar6!!"
        echo "MyPath1=$env:MyPath1!!"
        echo "MyPath2=$env:MyPath2!!"
        echo "MyPath3=$env:MyPath3!!"
        echo "MyPath4=$env:MyPath4!!"
    """)

    with chdir(temp_folder()):
        env = env.vars(ConanFileMock())
        env.save_ps1("test.ps1")
        save("display.ps1", display)
        cmd = "powershell.exe .\\test.ps1 ; .\\display.ps1 ; .\\deactivate_test.ps1 ; .\\display.ps1"
        check_env_files_output(cmd, prevenv)


@pytest.mark.skipif(platform.system() == "Windows", reason="Not in Windows")
def test_env_files_sh(env, prevenv):
    display = textwrap.dedent("""\
        echo MyVar=$MyVar!!
        echo MyVar1=$MyVar1!!
        echo MyVar2=$MyVar2!!
        echo MyVar3=$MyVar3!!
        echo MyVar4=$MyVar4!!
        echo MyVar5=$MyVar5!!
        echo MyVar6=$MyVar6!!
        echo MyPath1=$MyPath1!!
        echo MyPath2=$MyPath2!!
        echo MyPath3=$MyPath3!!
        echo MyPath4=$MyPath4!!
        """)

    with chdir(temp_folder()):
        env = env.vars(ConanFileMock())
        env.save_sh("test.sh")
        save("display.sh", display)
        os.chmod("display.sh", 0o777)
        # We include the "set -e" to test it is robust against errors
        cmd = 'set -e && . ./test.sh && ./display.sh && . ./deactivate_test.sh && ./display.sh'
        check_env_files_output(cmd, prevenv)



import unittest

from conans.model.package_metadata import PackageMetadata


class PackageMetadataTest(unittest.TestCase):

    def test_load_unload(self):
        a = PackageMetadata()
        a.recipe.revision = "rev"
        a.recipe.checksums["somefile1"] = {"md5": "50b2137a5d63567b7e88b743a3b594cf",
                                           "sha1": "0b7e8ed59ff4eacb95fd3cc8e17a8034584a96c2"}
        a.packages["ID"].recipe_revision = "rec_rev"
        a.packages["ID"].revision = "revp"
        a.packages["ID"].properties["Someprop"] = "23"
        a.packages["ID"].checksums["somefile2"] = {"md5": "efb7597b146344532fe8da2b79860aaa",
                                                   "sha1": "cc3e6eae41eca26538630f4cd5b0bf4fb52e2d"}

        tmp = a.dumps()

        b = PackageMetadata.loads(tmp)

        self.assertEqual(b, a)
        self.assertEqual(b.packages["ID"].properties["Someprop"], "23")
        self.assertEqual(b.recipe.checksums["somefile1"]["md5"],
                         "50b2137a5d63567b7e88b743a3b594cf")
        self.assertEqual(b.packages["ID"].checksums["somefile2"]["sha1"],
                         "cc3e6eae41eca26538630f4cd5b0bf4fb52e2d")

    def test_other_types_than_str(self):
        a = PackageMetadata()
        a.recipe.revision = "rev"
        a.packages["ID"].recipe_revision = 34
        a.packages["ID"].revision = {"23": 45}
        a.packages["ID"].properties["Someprop"] = [23, 2444]

        tmp = a.dumps()

        b = PackageMetadata.loads(tmp)

        self.assertEqual(b, a)
        self.assertEqual(b.packages["ID"].revision, {"23": 45})
        self.assertEqual(b.packages["ID"].properties["Someprop"], [23, 2444])



import os

from conan.tools.files import check_md5, check_sha1, check_sha256
from conans.test.utils.test_files import temp_folder
from conans.util.files import save


def test_checksums():
    tmp = temp_folder()
    file_path = os.path.join(tmp, "foo.txt")
    save(os.path.join(tmp, "foo.txt"), "contents")

    check_md5(None, file_path, "98bf7d8c15784f0a3d63204441e1e2aa")
    check_sha1(None, file_path, "4a756ca07e9487f482465a99e8286abc86ba4dc7")
    check_sha256(None, file_path, "d1b2a59fbea7e20077af9f91b27e95e865061b270be03ff539ab3b73587882e8")




import platform
import textwrap

import pytest

from conans.client.tools import environment_append
from conans.test.assets.cmake import gen_cmakelists
from conans.test.assets.sources import gen_function_cpp
from conans.test.utils.tools import TestClient
from conans.util.files import save


@pytest.fixture
def client():
    c = TestClient()
    save(c.cache.new_config_path, "tools.env.virtualenv:auto_use=True")
    clang_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        build_type=Release
        compiler=clang
        compiler.version=12
        """)
    conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile
        from conan.tools.cmake import CMake, cmake_layout

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = "*"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
                cmd = os.path.join(self.cpp.build.bindirs[0], "my_app")
                self.run(cmd, env=["conanrunenv"])
        """)
    c.save({"conanfile.py": conanfile,
            "clang": clang_profile,
            "CMakeLists.txt": gen_cmakelists(appname="my_app", appsources=["src/main.cpp"]),
            "src/main.cpp": gen_function_cpp(name="main")})
    return c


@pytest.mark.tool_cmake
@pytest.mark.tool_mingw64
@pytest.mark.tool_clang(version="12")
@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
def test_clang(client):
    client.run("create . pkg/0.1@ -pr=clang")
    # clang compilations in Windows will use MinGW Makefiles by default
    assert 'cmake -G "MinGW Makefiles"' in client.out
    assert "main __clang_major__12" in client.out
    # Check this! Clang compiler in Windows is reporting MSC_VER and MSVC_LANG!
    assert "main _MSC_VER19" in client.out
    assert "main _MSVC_LANG2014" in client.out


@pytest.mark.tool_cmake
@pytest.mark.tool_clang(version="12")
@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
def test_clang_cmake_ninja(client):
    client.run("create . pkg/0.1@ -pr=clang -c tools.cmake.cmaketoolchain:generator=Ninja")
    assert 'cmake -G "Ninja"' in client.out
    assert "main __clang_major__12" in client.out
    # Check this! Clang compiler in Windows is reporting MSC_VER and MSVC_LANG!
    assert "main _MSC_VER19" in client.out
    assert "main _MSVC_LANG2014" in client.out


@pytest.mark.tool_cmake
@pytest.mark.tool_clang(version="12")
@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
def test_clang_cmake_ninja_custom_cxx(client):
    with environment_append({"CXX": "/no/exist/clang++"}):
        client.run("create . pkg/0.1@ -pr=clang -c tools.cmake.cmaketoolchain:generator=Ninja",
                   assert_error=True)
        assert 'Could not find compiler' in client.out
        assert '/no/exist/clang++' in client.out

    clang_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        build_type=Release
        compiler=clang
        compiler.version=12
        [buildenv]
        CXX=/no/exist/clang++
        """)
    client.save({"clang":     clang_profile})
    client.run("create . pkg/0.1@ -pr=clang -c tools.cmake.cmaketoolchain:generator=Ninja",
               assert_error=True)
    assert 'Could not find compiler' in client.out
    assert '/no/exist/clang++' in client.out


@pytest.mark.tool_cmake
@pytest.mark.tool_visual_studio(version="16")  # With Clang distributed in VS!
@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
def test_clang_cmake_visual(client):
    clang_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        build_type=Release
        compiler=clang
        compiler.version=11
        """)
    # TODO: Clang version is unused, it can change, still 11 from inside VS is used
    client.save({"clang": clang_profile})
    client.run("create . pkg/0.1@ -pr=clang "
               '-c tools.cmake.cmaketoolchain:generator="Visual Studio 16"')
    assert 'cmake -G "Visual Studio 16"' in client.out
    assert "main __clang_major__11" in client.out
    # Check this! Clang compiler in Windows is reporting MSC_VER and MSVC_LANG!
    assert "main _MSC_VER19" in client.out
    assert "main _MSVC_LANG2014" in client.out



import os
import platform
import textwrap

import pytest

from conans.test.assets.genconanfile import GenConanfile
from conans.test.utils.tools import TestClient
from conans.tools import save


@pytest.fixture
def client():
    client = TestClient()
    conanfile = str(GenConanfile())
    conanfile += """
    def package_info(self):
        self.buildenv_info.define("Foo", "MyVar!")
        self.runenv_info.define("Foo", "MyRunVar!")
    """
    client.save({"conanfile.py": conanfile})
    client.run("create . foo/1.0@")
    save(client.cache.new_config_path, "tools.env.virtualenv:auto_use=True")
    return client


@pytest.mark.parametrize("default_virtualenv", [True, False, None])
def test_virtualenv_deactivated(client, default_virtualenv):
    format_str = {True: "virtualbuildenv = True",
                  False: "virtualbuildenv = False",
                  None: ""}[default_virtualenv]
    conanfile = textwrap.dedent("""
    from conans import ConanFile
    import platform

    class ConanFileToolsTest(ConanFile):
        {}
        requires = "foo/1.0"
    """).format(format_str)

    client.save({"conanfile.py": conanfile})
    client.run("install . ")
    extension = "bat" if platform.system() == "Windows" else "sh"
    exists_file = os.path.exists(os.path.join(client.current_folder,
                                              "conanbuildenv.{}".format(extension)))
    if default_virtualenv is True or default_virtualenv is None:
        assert exists_file
    elif default_virtualenv is False:
        assert not exists_file


def test_virtualrunenv_not_applied(client):
    """By default the VirtualRunEnv is not added to the list, otherwise when declaring
       generators = "VirtualBuildEnv", "VirtualRunEnv" will be always added"""
    conanfile = textwrap.dedent("""
    from conans import ConanFile
    import platform

    class ConanFileToolsTest(ConanFile):
        settings = "os"
        generators = "VirtualBuildEnv", "VirtualRunEnv"
        requires = "foo/1.0"
    """)

    client.save({"conanfile.py": conanfile})
    client.run("install . ")
    extension = "bat" if platform.system() == "Windows" else "sh"
    exists_file = os.path.exists(os.path.join(client.current_folder,
                                              "conanrun.{}".format(extension)))
    assert exists_file

    global_env = client.load("conanbuild.{}".format(extension))
    assert "conanrunenv" not in global_env


@pytest.mark.parametrize("explicit_declare", [True, False, None])
def test_virtualrunenv_explicit_declare(client, explicit_declare):
    """By default the VirtualRunEnv is not added to the list, otherwise when declaring
       generators = "VirtualBuildEnv", "VirtualRunEnv" will be always added"""
    conanfile = textwrap.dedent("""
    from conans import ConanFile
    from conan.tools.env import VirtualRunEnv
    import platform

    class ConanFileToolsTest(ConanFile):
        requires = "foo/1.0"

        def generate(self):
            VirtualRunEnv(self).generate({})

    """).format({True: "scope='build'",
                 False: "scope='run'",
                 None: ""}.get(explicit_declare))

    client.save({"conanfile.py": conanfile})
    client.run("install . ")
    extension = "bat" if platform.system() == "Windows" else "sh"
    exists_file = os.path.exists(os.path.join(client.current_folder,
                                              "conanbuild.{}".format(extension)))
    assert exists_file

    global_env = client.load("conanbuild.{}".format(extension))
    if explicit_declare:
        assert "conanrun" in global_env
    else:
        assert "conanrun" not in global_env



# coding=utf-8

import unittest

from conans.client.run_environment import RunEnvironment
from conans.test.utils.mocks import ConanFileMock


class RunEnvironmentTest(unittest.TestCase):

    def test_run_vars(self):
        conanfile = ConanFileMock()
        conanfile.deps_cpp_info["one"].bin_paths.append("path/bin")
        conanfile.deps_cpp_info["two"].lib_paths.append("path/libs")
        be = RunEnvironment(conanfile)

        self.assertEqual(be.vars,  {'PATH': ['path/bin'],
                                    'LD_LIBRARY_PATH': ['path/libs'],
                                    'DYLD_LIBRARY_PATH': ['path/libs']})

    def test_apple_frameworks(self):
        conanfile = ConanFileMock()
        conanfile.deps_cpp_info["one"].bin_paths.append("path/bin")
        conanfile.deps_cpp_info["two"].lib_paths.append("path/libs")
        conanfile.deps_cpp_info["one"].framework_paths.append("path/Frameworks")
        be = RunEnvironment(conanfile)

        self.assertEqual(be.vars, {'PATH': ['path/bin'],
                                   'LD_LIBRARY_PATH': ['path/libs'],
                                   'DYLD_LIBRARY_PATH': ['path/libs'],
                                   'DYLD_FRAMEWORK_PATH': ['path/Frameworks']})
